import datetime
import decimal
import logging
from base64 import b64decode, b64encode

from dateutil.parser import parse as dateparse
from sqlalchemy.types import String

log = logging.getLogger('unrest.coercers')


class Property(object):
    """
    A Property wrapper used instead of a string in a #::unrest.rest#Rest
    properties parameter.

    # Arguments
        name: This property name
        type: The sqlalchemy type used for type coercion
        formatter: An optional function to transform the parameter as string
    """

    def __init__(self, name, type=None, formatter=None):
        self.name = name
        self.type = type or String()
        self.formatter = formatter

    def get(self, serializer, model):
        prop = getattr(model, self.name)
        if self.formatter:
            return self.formatter(prop)
        else:
            prop = serializer._serialize(self.type, prop)
        return prop


class Serialize(object):
    """
    Base serializer class

    Casts python sqlalchemy data into a JSON compliant type according to
    the sqlalchemy column type.

    Not all types are implemented as of now and it's fairly easy to add:
    Just add a `serialize_type` method for `type` and it shall work.

    The serialize class can be configured with the rest function and
    on the UnRest declaration.

    For example:
    ```python
    from unrest.coercers import Serialize

    class BetterSerialize(Serialize):
        def serialize_matrix(self, type, data):
            return data.matrix_to_string()

    rest = UnRest(app, session, SerializeClass=BetterSerialize)
    ...
    ```

    # Arguments
        model: The sqlachemy item to serialize.
        columns: The list of columns to serialize.
        properties: The list of properties to serialize.
        relationships: The list of relationships to serialize.
    """

    def __init__(self, model, columns, properties, relationships):
        self.model = model
        self.columns = columns
        self.properties = properties
        self.relationships = relationships

    def dict(self):
        """Serialize the given model to a JSON compatible dict"""

        def enforce_iterable(it):
            try:
                return iter(it)
            except TypeError:
                return (it,)

        return dict(
            {
                name: self.serialize(name, column)
                for name, column in self.columns.items()
            },
            **dict(
                {
                    property.name: property.get(self, self.model)
                    for property in self.properties
                },
                **{
                    key: [
                        relationship_rest.serialize(item)
                        for item in enforce_iterable(getattr(self.model, key))
                    ]
                    for key, relationship_rest in self.relationships.items()
                },
            ),
        )

    def serialize(self, name, column):
        return self._serialize(column.type, getattr(self.model, name))

    def _serialize(self, type, data):
        if data is None:
            return
        method_name = f'serialize_{type.__class__.__name__.lower()}'

        if hasattr(self, method_name):
            return getattr(self, method_name)(type, data)

        log.debug(f'Missing method for type serialization {method_name}')

        return data

    def serialize_array(self, type, data):
        return [self._serialize(type.item_type, datum) for datum in data]

    def serialize_datetime(self, type, data):
        return data.isoformat()

    serialize_date = serialize_datetime
    serialize_time = serialize_datetime

    def serialize_interval(self, type, data):
        return data.total_seconds()

    def serialize_decimal(self, type, data):
        return float(data)

    serialize_numeric = serialize_decimal
    serialize_float = serialize_decimal

    def serialize_largebinary(self, type, data):
        return b64encode(data).decode('utf-8')


class Deserialize(object):
    """
    Base deserializer class

    Casts raw data back to compatible python sqlalchemy type.

    Not all types are implemented as of now and it's fairly easy to add:
    Just add a `deserialize_type` method for `type` and it shall work.

    The deserialize class can be configured with the rest function and
    on the UnRest declaration.

    For example:
    ```python
    from unrest.coercers import Deserialize

    class BetterDeserialize(Deserialize):
        def deserialize_matrix(self, type, data):
            return Matrix.from_string(data)

    rest = UnRest(app, session, DeserializeClass=BetterDeserialize)
    ...
    ```

    # Arguments
        payload: The payload to deserialize
        columns: The list of columns to deserialize
    """

    def __init__(self, payload, columns):
        self.payload = payload
        self.columns = columns

    def merge(self, item, payload=None):
        """Deserialize the given payload into the existing sqlalchemy `item`"""
        for name, column in self.columns.items():
            setattr(item, name, self.deserialize(name, column, payload))
        return item

    def create(self, factory):
        """
        Deserialize objects in the given payload into a list of new items
        created with the `factory` function.
        """
        return [
            self.merge(factory(), item) for item in self.payload['objects']
        ]

    def deserialize(self, name, column, payload=None):
        payload = payload or self.payload
        if name not in payload:
            return None
        return self._deserialize(column.type, payload[name])

    def _deserialize(self, type, data):
        if data is None:
            return
        method_name = f'deserialize_{type.__class__.__name__.lower()}'

        if hasattr(self, method_name):
            return getattr(self, method_name)(type, data)

        log.debug(f'Missing method for type deserialization {method_name}')

        return data

    def deserialize_array(self, type, data):
        return [self._deserialize(type.item_type, datum) for datum in data]

    def deserialize_datetime(self, type, data):
        return dateparse(data)

    def deserialize_date(self, type, data):
        return dateparse(data).date()

    def deserialize_time(self, type, data):
        return dateparse(data).time()

    def deserialize_interval(self, type, data):
        return datetime.timedelta(seconds=data)

    def deserialize_integer(self, type, data):
        return int(data)

    def deserialize_decimal(self, type, data):
        return decimal.Decimal(data)

    deserialize_numeric = deserialize_decimal
    deserialize_float = deserialize_decimal

    def deserialize_largebinary(self, type, data):
        return b64decode(data)



import sys

from pytest import raises
from sqlalchemy.dialects.postgresql import INET
from sqlalchemy.types import Boolean, Float, String

from unrest import UnRest, __about__
from unrest.coercers import Deserialize, Serialize
from unrest.rest import Rest

from .. import idsorted
from ...framework import Framework
from ...framework.flask import FlaskFramework
from ..model import Fruit, Tree
from ..static.openapi_result import openapi


def test_other_url(client):
    rest = UnRest(client.app, client.session, framework=client.__framework__)
    rest(Tree)
    code, html = client.fetch('/')
    assert code == 200
    assert html == 'A normal route!'


def test_index(client):
    rest = UnRest(client.app, client.session, framework=client.__framework__)
    rest(Tree)
    code, html = client.fetch('/api/')
    assert code == 200
    assert html == (
        '<h1>unrest <small>api server</small></h1> version '
        f'{__about__.__version__} '
        '<a href="https://github.com/Kozea/unrest">unrest</a> '
        '<a href="/api/openapi.json">openapi.json</a>'
    )


def test_normal(client):
    rest = UnRest(client.app, client.session, framework=client.__framework__)
    rest(Tree)
    code, json = client.fetch('/api/tree')
    assert code == 200
    assert json['occurences'] == 3


def test_path(client):
    rest = UnRest(
        client.app, client.session, '/forest', framework=client.__framework__
    )
    rest(Tree)
    code, json = client.fetch('/forest/tree')
    assert code == 200
    assert json['occurences'] == 3


def test_schema(client):
    Tree.__table__.schema = 'forest'

    def reset_schema(query):
        Tree.__table__.schema = None
        return query

    rest = UnRest(client.app, client.session, framework=client.__framework__)
    rest(Tree, query=reset_schema)
    code, json = client.fetch('/api/forest/tree')
    assert code == 200
    assert json['occurences'] == 3


def test_version(client):
    rest = UnRest(
        client.app,
        client.session,
        version='v3.14',
        framework=client.__framework__,
    )
    rest(Tree)
    code, json = client.fetch('/api/v3.14/tree')
    assert code == 200
    assert json['occurences'] == 3


def test_path_and_version(client):
    rest = UnRest(
        client.app,
        client.session,
        '/forest',
        'v3.14',
        framework=client.__framework__,
    )
    rest(Tree)
    code, json = client.fetch('/forest/v3.14/tree')
    assert code == 200
    assert json['occurences'] == 3


def test_normal_rest_class(client):
    rest = UnRest(client.app, client.session, framework=client.__framework__)
    tree = rest(Tree, name='tree')
    assert isinstance(tree, Rest)


def test_alternative_rest_class(client):
    class NewRest(Rest):
        def __init__(client, *args, **kwargs):
            kwargs['name'] = 'new_' + kwargs['name']
            super().__init__(*args, **kwargs)

    new_rest = UnRest(
        client.app,
        client.session,
        framework=client.__framework__,
        RestClass=NewRest,
    )
    new_tree = new_rest(Tree, name='tree')
    assert isinstance(new_tree, NewRest)

    code, json = client.fetch('/api/tree')
    assert code == 404
    code, json = client.fetch('/api/new_tree')
    assert code == 200
    assert json['occurences'] == 3


def test_empty_get_pk_as_404(client):
    rest = UnRest(
        client.app,
        client.session,
        framework=client.__framework__,
        empty_get_as_404=True,
    )
    rest(Tree)
    code, json = client.fetch('/api/tree/6')
    assert code == 404
    assert json['occurences'] == 0
    assert json['objects'] == []


def test_bad_json(client):
    rest = UnRest(client.app, client.session, framework=client.__framework__)
    rest(Tree, methods=['GET', 'POST'])
    code, json = client.fetch(
        '/api/tree', method="POST", body="{'name'; 'cedar'}"
    )
    assert code == 400
    assert (
        json['message'] == 'JSON Error in payload: '
        'Expecting property name enclosed in double quotes: '
        'line 1 column 2 (char 1)'
    )


def test_empty_explicit_framework(client):
    class FakeUnRest(object):
        def __init__(client, app, url):
            pass

        def register_route(client, *args, **kwargs):
            pass

    rest = UnRest(client.app, client.session, framework=FakeUnRest)
    rest(Tree)
    code, _ = client.fetch('/api/tree')
    assert code == 404


def test_api_options(client):
    rest = UnRest(client.app, client.session, framework=client.__framework__)
    fruit = rest(Fruit)
    rest(
        Tree,
        methods=rest.all,
        relationships={'fruits': fruit},
        properties=['fruit_colors'],
        allow_batch=True,
    )
    code, json = client.fetch('/api', method="OPTIONS")
    assert code == 200
    assert json == {
        '/api/fruit': {
            'model': 'Fruit',
            'description': 'A bag of fruit',
            'parameters': ['fruit_id'],
            'columns': {
                'age': 'timedelta',
                'color': 'str',
                'fruit_id': 'int',
                'size': 'Decimal',
                'tree_id': 'int',
                'double_size': 'Decimal',
            },
            'properties': {},
            'relationships': {},
            'methods': ['GET', 'OPTIONS'],
            'batch': False,
        },
        '/api/tree': {
            'model': 'Tree',
            'description': "Where money doesn't grow",
            'parameters': ['id'],
            'columns': {'id': 'int', 'name': 'str'},
            'properties': {'fruit_colors': 'str'},
            'relationships': {
                'fruits': {
                    'model': 'Fruit',
                    'description': 'A bag of fruit',
                    'parameters': ['fruit_id'],
                    'columns': {
                        'age': 'timedelta',
                        'color': 'str',
                        'fruit_id': 'int',
                        'size': 'Decimal',
                        'tree_id': 'int',
                        'double_size': 'Decimal',
                    },
                    'properties': {},
                    'relationships': {},
                    'batch': False,
                }
            },
            'methods': ['GET', 'PUT', 'POST', 'DELETE', 'PATCH', 'OPTIONS'],
            'batch': True,
        },
    }


def test_endpoint_options(client):
    rest = UnRest(client.app, client.session, framework=client.__framework__)
    fruit = rest(Fruit)
    rest(
        Tree,
        methods=rest.all,
        relationships={'fruits': fruit},
        properties=['fruit_colors'],
        allow_batch=True,
    )
    code, json = client.fetch('/api/fruit', method="OPTIONS")
    assert code == 200
    assert json == {
        'model': 'Fruit',
        'description': 'A bag of fruit',
        'batch': False,
        'parameters': ['fruit_id'],
        'columns': {
            'age': 'timedelta',
            'color': 'str',
            'fruit_id': 'int',
            'size': 'Decimal',
            'tree_id': 'int',
            'double_size': 'Decimal',
        },
        'properties': {},
        'relationships': {},
        'methods': ['GET', 'OPTIONS'],
    }


def test_endpoint_options_other_type(client):
    rest = UnRest(client.app, client.session, framework=client.__framework__)
    rest(
        Tree,
        properties=[
            rest.Property('fake_prop_1', type=Boolean()),
            rest.Property('fake_prop_2', type=INET()),
        ],
    )
    code, json = client.fetch('/api/tree', method="OPTIONS")
    assert code == 200
    assert {
        'model': 'Tree',
        'description': "Where money doesn't grow",
        'parameters': ['id'],
        'columns': {'id': 'int', 'name': 'str'},
        'properties': {'fake_prop_1': 'bool', 'fake_prop_2': 'INET'},
        'relationships': {},
        'methods': ['GET', 'OPTIONS'],
        'batch': False,
    }


def test_endpoint_options_no_methods(client):
    rest = UnRest(client.app, client.session, framework=client.__framework__)
    rest(Tree, methods=[])
    code, json = client.fetch('/api/tree', method="OPTIONS")
    assert code == 404


def test_endpoint_options_no_methods_but_declare(client):
    rest = UnRest(client.app, client.session, framework=client.__framework__)
    tree = rest(Tree, methods=[])

    @tree.declare('GET')
    def get(payload, id=None):
        return {'Hey': 'Overridden'}

    @tree.declare('POST')
    def post(payload, id=None):
        return {'Hey': 'Overridden'}

    code, json = client.fetch('/api/tree', method="GET")
    assert code == 200
    code, json = client.fetch('/api/tree', method="POST", json={})
    assert code == 200

    code, json = client.fetch('/api/tree', method="OPTIONS")
    assert code == 200


def test_openapi(client):
    rest = UnRest(
        client.app,
        client.session,
        info={
            'description': '''# Unrest demo
This is the demo of unrest api.
This api expose the `Tree` and `Fruit` entity Rest methods.
''',
            'contact': {
                'name': __about__.__author__,
                'url': __about__.__uri__,
                'email': __about__.__email__,
            },
            'license': {'name': __about__.__license__},
        },
        framework=client.__framework__,
    )
    fruit = rest(
        Fruit,
        methods=rest.all,
        properties=[rest.Property('square_size', Float())],
    )
    rest(
        Tree,
        methods=rest.all,
        relationships={'fruits': fruit},
        properties=['fruit_colors'],
        allow_batch=True,
    )
    code, json = client.fetch('/api/openapi.json')
    # Only flask supports absolute url
    if 'Flask' not in client.__framework__.__name__:
        res = {**openapi, 'servers': [{'url': '/api'}]}
    else:
        res = openapi
    assert code == 200
    assert json == res


def test_openapi_other_type(client):
    rest = UnRest(client.app, client.session, framework=client.__framework__)
    rest(
        Tree,
        properties=[
            rest.Property('fake_prop_1', type=Boolean()),
            rest.Property('fake_prop_2', type=INET()),
        ],
    )
    code, json = client.fetch('/api/openapi.json')
    assert code == 200
    assert json['paths']['/tree']['get']['responses']['200']['content'][
        'application/json'
    ]['schema']['properties']['objects']['items']['properties'] == {
        'fake_prop_1': {'type': 'boolean'},
        'fake_prop_2': {'type': 'string'},
        'id': {'format': 'int64', 'type': 'integer'},
        'name': {'type': 'string'},
    }


def test_sub(client):
    rest = UnRest(client.app, client.session, framework=client.__framework__)

    fruit = rest(
        Fruit,
        methods=rest.all,
        properties=[rest.Property('square_size', Float())],
    )
    tree = rest(
        Tree,
        methods=rest.all,
        relationships={'fruits': fruit},
        properties=['fruit_colors'],
        query=lambda q: q.filter(Tree.name != 'pine'),
        allow_batch=True,
    )
    subtree = tree.sub(lambda q: q.filter(Tree.name != 'oak'))
    for key in [
        'unrest',
        'Model',
        'methods',
        'only',
        'exclude',
        'properties',
        'relationships',
        'allow_batch',
        'auth',
        'read_auth',
        'write_auth',
        'validators',
        '_primary_keys',
        'SerializeClass',
        'DeserializeClass',
    ]:
        assert getattr(subtree, key) == getattr(tree, key)
    assert subtree.name == 'subtree'

    code, json = client.fetch('/api/tree')
    assert code == 200
    assert json['occurences'] == 2

    code, json = client.fetch('/api/subtree')
    assert code == 200
    assert json['occurences'] == 1

    subtree = tree.sub(
        lambda q: q.filter(Tree.name != 'maple'), name='nomaple'
    )
    code, json = client.fetch('/api/nomaple')
    assert code == 200
    assert json['occurences'] == 1


def test_sub_fixed(client):
    rest = UnRest(client.app, client.session, framework=client.__framework__)

    fruit = rest(
        Fruit,
        defaults={'size': 1.0},
        fixed={'color': 'blue'},
        methods=rest.all,
        properties=[rest.Property('square_size', Float())],
    )
    subfruit = fruit.sub(lambda q: q.filter(Fruit.age == 2.0))
    for key in ['defaults', 'fixed']:
        assert getattr(subfruit, key) == getattr(fruit, key)


def test_wrong_framework(client):
    with raises(NotImplementedError):
        UnRest(client.app, client.session, framework=Framework)


def test_auto_framework(client):
    if 'Flask' not in client.__framework__.__name__:
        return
    rest = UnRest(client.app, client.session)
    assert isinstance(rest.framework, FlaskFramework)


def test_no_framework(client):
    if 'Flask' not in client.__framework__.__name__:
        return
    flask = sys.modules['flask']
    sys.modules['flask'] = None

    with raises(NotImplementedError):
        UnRest(client.app, client.session)

    sys.modules['flask'] = flask


def test_custom_serialization(client):
    class UpperCaseStringSerialize(Serialize):
        def serialize(client, name, column):
            rv = super().serialize(name, column)
            if isinstance(column.type, String):
                return rv.upper()
            return rv

    rest = UnRest(
        client.app,
        client.session,
        SerializeClass=UpperCaseStringSerialize,
        framework=client.__framework__,
    )
    rest(Tree, methods=['GET', 'PUT'])

    code, json = client.fetch('/api/tree')
    assert code == 200
    assert json['occurences'] == 3
    assert idsorted(json['objects']) == [
        {'id': 1, 'name': 'PINE'},
        {'id': 2, 'name': 'MAPLE'},
        {'id': 3, 'name': 'OAK'},
    ]


def test_custom_deserialization(client):
    class UpperCaseStringDeserialize(Deserialize):
        def deserialize(client, name, column, payload=None):
            rv = super().deserialize(name, column, payload)
            if isinstance(column.type, String):
                return rv.upper()
            return rv

    rest = UnRest(
        client.app,
        client.session,
        DeserializeClass=UpperCaseStringDeserialize,
        framework=client.__framework__,
    )
    rest(Tree, methods=['GET', 'PUT'])

    code, json = client.fetch(
        '/api/tree/1', method="PUT", json={'id': 1, 'name': 'cedar'}
    )
    assert code == 200
    assert json['occurences'] == 1
    assert idsorted(json['objects']) == [{'id': 1, 'name': 'CEDAR'}]

    code, json = client.fetch('/api/tree')
    assert code == 200
    assert json['occurences'] == 3
    assert idsorted(json['objects']) == [
        {'id': 1, 'name': 'CEDAR'},
        {'id': 2, 'name': 'maple'},
        {'id': 3, 'name': 'oak'},
    ]


def test_wrong_method(client):
    rest = UnRest(client.app, client.session, framework=client.__framework__)
    rest(Tree, methods=['GET', 'POST'])
    code, html = client.fetch(
        '/api/tree',
        method="PUT",
        json={
            'objects': [{'id': 1, 'name': 'cedar'}, {'id': 2, 'name': 'mango'}]
        },
    )
    assert code == 405


def test_wrong_url(client):
    rest = UnRest(client.app, client.session, framework=client.__framework__)
    rest(Tree)
    code, html = client.fetch('/apy/trea')
    assert code == 404


def test_duplicate_rest(client):
    rest = UnRest(client.app, client.session, framework=client.__framework__)
    rest(Tree)

    with raises(Exception):
        rest(Tree)



from sanic import Sanic, response

from ...framework.sanic import SanicFramework
from .unrest_client import UnRestClient


class SanicClient(UnRestClient):
    __framework__ = SanicFramework

    def setUp(self):
        self.get_app()
        super().setUp()

    def get_app(self):
        self.app = Sanic()

        @self.app.route("/")
        async def home(request):
            return response.text("A normal route!")

        @self.app.middleware('response')
        async def after_request(request, response):
            self.app.session.remove()

        self.app.session = self.session
        return self.app

    def raw_fetch(self, url, method='GET', headers={}, body=None):
        req, res = getattr(self.app.test_client, method.lower())(
            url, headers=headers, data=body
        )
        res.code = res.status
        return res



import logging
from contextlib import contextmanager
from functools import partial

from sqlalchemy import and_, or_
from sqlalchemy.inspection import inspect
from sqlalchemy.orm.query import Query
from sqlalchemy.orm.strategy_options import Load

from .coercers import Deserialize, Serialize
from .generators.options import Options
from .idiom.unrest import UnRestIdiom

log = logging.getLogger(__name__)


def _call_me_maybe(fun_or_value, *args, **kwargs):
    """Call first argument `fun_or_value` with `*args, **kwargs`
        if it is callable, return it otherwise.
    """
    if callable(fun_or_value):
        return fun_or_value(*args, **kwargs)
    return fun_or_value


def _identity(arg):
    """Identity function, return `arg`"""
    return arg


class Rest(object):
    """
    This is the entry point for generating a REST endpoint for a specific model
    The final uri if the path is '/api' and version 'v2' would be:
    `/api/v2/model` and `/api/v2/model/pk1/pk2` and if model is not in the
    public schema `/api/v2/schema/model` and `/api/v2/schema/model/pk1/pk2`.

    Usage:
    ```python
        rest = UnRest(app)

        def name_validator(field):
            if len(field.value) > 12:
                raise field.ValidationError(
                    'Name is too long (max 12 characters).')
            return field.value

        rest(Person, only=['name', 'sex', 'age'], methods=rest.all,
             query=lambda q: q.filter(Person.age > 16),
             validators={'name': name_validator})
    ```
    # Arguments
        unrest: The unrest instance given automatically on UnRest call.
        Model: The sqlalchemy orm model class.
        methods: The allowed method list on this endpoint. Possible values are
            GET, PUT, POST, DELETE, PATCH and rest.all
        name: If specified replaces the model name in url.
        only: If specified restricts the json fields to this list.
        exclude: If specified removes the json fields in this list.
        query: A function that takes the Model query and returns your specific
            query. Can be useful to filter data for all the methods.
        properties: A list of additional properties to retrieve on the model.
        relationships: A mapping of relationships and rest endpoints to fetch
            with the model.
        allow_batch: Allow batch operations (PUT, DELETE and PATCH)
            without primary key.
        auth: A decorator that will always be called.
        read_auth: A decorator that will be called on GET.
        write_auth: A decorator that will be called on PUT, POST, DELETE
            and PATCH.
        validators: A mapping of field names and validation functions.
            A validator function takes a #::unrest.rest.Rest#Validatable
            object as parameter and must return the final value for the field
            or raise a `rest.ValidationError(reason)` (where `rest = Unrest()`)
        validation_error_code: The http return code when the validation fails.
            Defaults to 500
        primary_keys: A list of column names to use as primary_keys
            (use real db primary keys by default)
        defaults: A mapping of column -> values which sets the default value
            of the columns when the column is not present in the payload.
            Can be a callable, in this case it will be called at runtime with
            the payload as argument.
        fixed: A mapping of column -> values which replaces the values
            present or not in the payload. Can be a callable, in this case it
            will be called at runtime with the payload as argument.
        SerializeClass: An alternative #::unrest.coercers#Serialize class.
        DeserializeClass: An alternative #::unrest.coercers#Deserialize class.
    """

    def __init__(
        self,
        unrest,
        Model,
        methods=['GET'],
        name=None,
        only=None,
        exclude=None,
        query=None,
        properties=None,
        relationships=None,
        allow_batch=False,
        auth=None,
        read_auth=None,
        write_auth=None,
        validators=None,
        validation_error_code=500,
        primary_keys=None,
        defaults=None,
        fixed=None,
        idiom=UnRestIdiom,
        SerializeClass=Serialize,
        DeserializeClass=Deserialize,
    ):
        self.unrest = unrest
        self.unrest.rests.append(self)
        self.Model = Model

        self.methods = tuple(methods)
        self.name = name or self.table.name
        self.only = only
        self.exclude = exclude
        self.query_factory = query or (lambda q: q)
        self.properties = [
            self.unrest.Property(property)
            if not isinstance(property, self.unrest.Property)
            else property
            for property in (properties or [])
        ]
        self.relationships = relationships or {}

        self.allow_batch = allow_batch

        self.auth = auth
        self.read_auth = read_auth
        self.write_auth = write_auth

        self.validators = validators or {}
        self.validation_error_code = validation_error_code
        self._primary_keys = primary_keys
        self.defaults = defaults or {}
        self.fixed = fixed or {}

        self.idiom = idiom(self)

        self.SerializeClass = SerializeClass
        self.DeserializeClass = DeserializeClass

        self._query_alterer = _identity

        self.overrides = {}

        for method in self.methods:
            self.register_method(method)

    def get(self, payload, **pks):
        """
        The GET method

        - With no arguments: Returns all query elements. (/api/model/)
        - With primary keys: Returns the element in query with the primary
            keys. (/api/model/pk)

        # Arguments
            payload: The request content ignored for GET.
            pks: The primary keys in url if any.
        """
        if self.has(pks):
            item = self.get_from_pk(self.query, **pks)
            return self.serialize_all([item] if item else [])

        items = self.query
        return self.serialize_all(items)

    def put(self, payload, **pks):
        """
        The PUT method

        - With no arguments: If allow_batch set to true replace all the
            query elements with the ones in the request payload.
        - With primary keys: Create or replace the element associated
            with the primary keys from the one in the request payload.

        # Arguments
            payload: The request content containing new elements.
            pks: The primary keys in url if any.
        """
        if not payload:
            self.raise_error(400, 'You must provide a payload')

        if self.has(pks):
            for pk, val in pks.items():
                if pk in payload:
                    assert payload[pk] == val, (
                        f'Incoherent primary_key ({pk}) in payload '
                        f'({payload[pk]!r}) and url ({val!r}) for PUT'
                    )
                else:
                    payload[pk] = val
            existingItem = self.get_from_pk(self.query, **pks)
            item = self.deserialize(payload, existingItem or self.Model())
            self.validate(item)
            if existingItem is None:
                self.session.add(item)
            self.session.flush()
            self.session.expire(item)
            return self.serialize_all([item])

        if not self.allow_batch:
            raise self.unrest.RestError(
                406,
                'You must set allow_batch to True '
                'if you want to use batch methods.',
            )

        self.query.delete()
        items = self.deserialize_all(payload)
        self.validate_all(items)
        self.session.add_all(items)
        self.session.flush()
        self.session.expire_all()
        return self.serialize_all(items)

    def post(self, payload, **pks):
        """
        The POST method

        - With no arguments: Add element from request payload.
        - With primary keys: Correspond to new collection creation. Unused.

        # Arguments
            payload: The request content containing the new element.
            pks: The primary keys in url if any.
        """
        if self.has(pks):
            # Create a collection?
            raise self.unrest.RestError(
                501,
                "POST with primary keys corresponds to collection "
                "creation. It's not implemented by default. "
                "If you want to update an item use the PUT method instead",
            )

        if not payload:
            self.raise_error(400, 'You must provide a payload')
        item = self.deserialize(payload, self.Model())
        self.session.add(item)
        self.validate(item)
        self.session.flush()
        self.session.expire(item)
        return self.serialize_all([item])

    def delete(self, payload, **pks):
        """
        The DELETE method

        - With no arguments: If allow_batch set to true delete all query
            elements.
        - With primary keys: Delete the element associated with the primary
            keys.

        # Arguments
            payload: The request content ignored in DELETE.
            pks: The primary keys of the element to delete.
        """
        if self.has(pks):
            item = self.get_from_pk(self.undefered_query, **pks)
            if item is None:
                self.raise_error(404, f'{self.name}({pks!r}) not found')

            self.session.delete(item)
            self.session.flush()
            return self.serialize_all([item])

        if not self.allow_batch:
            raise self.unrest.RestError(
                406,
                'You must set allow_batch to True '
                'if you want to use batch methods.',
            )

        items = self.undefered_query.all()
        self.query.delete()
        self.session.flush()
        return self.serialize_all(items)

    def patch(self, payload, **pks):
        """
        The PATCH method

        - With no arguments: If allow_batch set to true patch existing elements
            with element attributes specified in the request payload.
        - With primary keys: Patch only one

        # Arguments
            payload: The request content containing
                a list of attributes to be patched.
            pks: The primary keys of the element to patch.
        """
        if not payload:
            self.raise_error(400, 'You must provide a payload')

        if self.has(pks):
            for pk, val in pks.items():
                if pk in payload:
                    assert payload[pk] == val, (
                        f'Incoherent primary_key ({pk}) in payload '
                        f'({payload[pk]!r}) and url ({val!r}) for PATCH'
                    )
                else:
                    payload[pk] = val
            item = self.get_from_pk(self.query, **pks)
            if item is None:
                self.raise_error(404, f'{self.name}({pks!r}) not found')
            self.deserialize(payload, item, blank_missing=False)
            self.validate(item)
            self.session.flush()
            self.session.expire(item)
            return self.serialize_all([item])

        if not self.allow_batch:
            raise self.unrest.RestError(
                406,
                'You must set allow_batch to True '
                'if you want to use batch methods.',
            )

        patches = payload['objects']
        # Get all concerned items
        items = self.get_all_from_pks(
            self.query,
            [{pk: patch[pk] for pk in self.primary_keys} for patch in patches],
        )
        if len(items) < len(patches):
            for patch in patches:
                if (
                    len(
                        [
                            it
                            for pk in self.primary_keys
                            for it in items
                            if getattr(it, pk) == patch[pk]
                        ]
                    )
                    == 0
                ):
                    patch = {
                        key: val
                        for key, val in patch.items()
                        if key in self.primary_keys
                    }
                    self.raise_error(404, f'{self.name}({patch}) not found')

        for patch in patches:
            # Get the patch item
            item = [
                it
                for pk in self.primary_keys
                for it in items
                if getattr(it, pk) == patch[pk]
            ][0]
            # Merge only patched colmuns
            self.deserialize(patch, item, blank_missing=False)
        self.validate_all(items)
        self.session.flush()
        self.session.expire_all()
        return self.serialize_all(items)

    def options(self, payload, **pks):
        """
        The OPTIONS method

        Returns a description of this rest endpoint.
        """
        return Options(self.unrest).get_route(self)

    def declare(self, method, manual_commit=False):
        """
        A decorator to register an alternative method.
        The original is still callable with rest.{method}

        ```python
        fruit = rest(Fruit)

        @fruit.declare('GET')
        def get(payload, fruit_id=None):
            rv = fruit.get(payload, fruit_id=fruit_id)
            return {
                'occurences': rv['occurences'],
                'objects': [
                    {'id': obj['fruit_id']} for obj in rv['objects']
                ]
            }
        ```

        # Arguments
            method: The method to override ('GET' for exemple)
            manual_commit: Set this to True to prevent auto commit after route
                call
        """

        def register_function(function):
            self.overrides[method] = (function, manual_commit)
            if method not in self.methods:
                self.register_method(method)
            return function

        return register_function

    def sub(self, query_factory, **kwargs):
        """
        This methods return a copy of the current rest endpoint and takes a
        `query_factory` argument to alter the current query.

        # Arguments
            query_factory: A function that takes the original query
                in parameter and returns a new query.
            **kwargs: Can be used to override Rest constructor arguments
                (query is not supported)

        # Returns
        A #::unrest.rest#Rest endpoint copied from this one
        """

        assert not kwargs.get('query'), 'query is not supported on sub rest'
        inherited = {
            'methods': self.methods,
            'name': 'sub' + self.name,
            'only': self.only,
            'exclude': self.exclude,
            'properties': self.properties,
            'relationships': self.relationships,
            'allow_batch': self.allow_batch,
            'auth': self.auth,
            'read_auth': self.read_auth,
            'write_auth': self.write_auth,
            'validators': self.validators,
            'primary_keys': self._primary_keys,
            'SerializeClass': self.SerializeClass,
            'DeserializeClass': self.DeserializeClass,
            'fixed': self.fixed,
            'defaults': self.defaults,
        }
        inherited.update(kwargs)
        subrest = self.__class__(self.unrest, self.Model, **inherited)
        subrest.query_factory = lambda q: query_factory(self.query_factory(q))
        return subrest

    def parameters_to_pks(self, parameters):
        """
        Transform query parameters into primary keys mapping with
        deserialized values.
        """
        if not parameters or all(
            value is None for value in parameters.values()
        ):
            return {}

        # In case of column_property or hybrid_property
        prop_by_name = {prop.name: prop for prop in self.properties}
        columns = {
            pk: self.columns.get(pk, prop_by_name.get(pk))
            for pk in self.primary_keys
        }
        deserialize = self.DeserializeClass(parameters, columns)
        return {
            name: deserialize.deserialize(name, column)
            for name, column in columns.items()
        }

    def deserialize(self, payload, item, blank_missing=True):
        """
        Deserialize the payload item in the provided item.

        # Arguments
            payload: The payload containing the item object
            item: An instance of the model to put values in
            blank_missing: Set non-provided by payload item attributes at None
        """

        if blank_missing:
            columns = self.columns
        else:
            # Mind only provided columns
            columns = {
                name: column
                for name, column in self.columns.items()
                if name in payload
            }
        self.set_defaults(payload, columns)
        return self.DeserializeClass(payload, columns).merge(item)

    def deserialize_all(self, payload):
        """
        Deserialize all the payload items.

        # Arguments
            payload: The payload containing the item list
        """
        for item in payload['objects']:
            self.set_defaults(item, self.columns)
        return self.DeserializeClass(payload, self.columns).create(self.Model)

    def serialize(self, item):
        """Serialize an `item` with the given `SerializeClass`"""
        return self.SerializeClass(
            item, self.columns, self.properties, self.relationships
        ).dict()

    def serialize_all(self, items):
        """
        Serialize all items and return a mapping containing:

        # Returns
        A dict containing:

        - objects: The serialized objects
        - primary_keys: The list of primary keys defined for this rest
            endpoint
        - occurences: The number of total occurences (without limit)
        - offset if there's a query offset
        - limit if there's a query limit
        """

        rv = {}
        rv['primary_keys'] = self.primary_keys

        if isinstance(items, Query):
            rv['occurences'] = items.offset(None).limit(None).count()
            if items._offset is not None:
                rv['offset'] = items._offset
            if items._limit is not None:
                rv['limit'] = items._limit

        rv['objects'] = [self.serialize(item) for item in items]
        if 'occurences' not in rv:
            rv['occurences'] = len(rv['objects'])
        return rv

    def set_defaults(self, payload, columns):
        """Sets in payload item all the fixed and defaults values"""
        for name, column in columns.items():
            if name in self.fixed:
                payload[name] = _call_me_maybe(self.fixed[name], payload)
            elif name not in payload and name in self.defaults:
                payload[name] = _call_me_maybe(self.defaults[name], payload)

    class Validatable(object):
        """
        A validatable class that is used as validators argument.

        # Arguments
            value: The current field value
            name: The current field name
            item: The current item
            ValidationError: The #::unrest.UnRest#ValidationError Exception
                to raise on validation error.
        """

        def __init__(self, value, name, item, ValidationError):
            self.value = value
            self.name = name
            self.item = item
            self.ValidationError = ValidationError

    def validate(self, item, errors=None):
        """
        Validates all validators columns against validators.

        Raise RestError if validation errors.
        """
        with self.session.no_autoflush:
            valid = True
            error = {'fields': {}}
            for pk in self.primary_keys:
                error[pk] = getattr(item, pk)

            for key, validators in self.validators.items():
                field_errors = []
                if callable(validators):
                    validators = (validators,)
                try:
                    for validator in validators:
                        setattr(
                            item,
                            key,
                            validator(
                                self.Validatable(
                                    getattr(item, key),
                                    key,
                                    item,
                                    self.unrest.ValidationError,
                                )
                            ),
                        )
                except self.unrest.ValidationError as e:
                    valid = False
                    field_errors.append(e.message)
                if field_errors:
                    error['fields'][key] = '\n'.join(field_errors)
            if errors is not None:
                errors.append(error)
            elif not valid:
                self.raise_error(
                    self.validation_error_code,
                    'Validation Error',
                    extra={'errors': [error]},
                )

    def validate_all(self, items):
        """
        Calls validate on all the `items` and raise an error if any does not
        validate.

        # Arguments
            items: A list of item to validate

        # Raises
        A #::unrest.UnRest#RestError on validation error
        """

        errors = []
        for item in items:
            self.validate(item, errors=errors)
        if any(len(error['fields']) for error in errors):
            self.raise_error(500, 'Validation Error', extra={'errors': errors})

    def raise_error(self, status, message, extra=None):
        """Shortcut function to #::unrest.UnRest#raise_error."""
        self.unrest.raise_error(status, message, extra)

    def route(self, method, request):
        """
        This is the entry point for any route method ( #get, #post, #put,
        #delete, #patch, #options ), it takes the #::unrest.util#Request and
        returns the #::unrest.util#Response.

        Exhaustively it
        - converts request parameters into primary keys values
        - calls the current idiom #::unrest.idiom.Idiom#request_to_payload
        - checks auth with `read_auth`, `write_auth` and `auth` if defined
        - calls the route associated with the HTTP `method`, either by default
            or overidden with #declare, with the previously obtained payload
        - commits the session if `manual_commit` is `False` and method is
            amongst modification ones
        - and finally calls #::unrest.idiom.Idiom#data_to_response with the
            return value of the wrapped function to return
            the #::unrest.util#Response

        # Arguments
            method: The HTTP method which is curried in a partial
            request: The current #::unrest.util#Request

        # Returns
        The #::unrest.util#Response of this request
        """

        try:
            pks = self.parameters_to_pks(request.parameters)
            payload = self.idiom.request_to_payload(request)

            route = getattr(self, method.lower())
            manual_commit = False
            if method in self.overrides:
                route, manual_commit = self.overrides[method]

            if method == 'GET' and self.read_auth:
                route = self.read_auth(route)
            if (
                method in ['PUT', 'POST', 'DELETE', 'PATCH']
                and self.write_auth
            ):
                route = self.write_auth(route)
            if self.auth:
                route = self.auth(route)

            with self.query_request(request):
                data = route(payload, **pks)

            if not manual_commit and method in [
                'PUT',
                'POST',
                'DELETE',
                'PATCH',
            ]:
                self.session.commit()
        except self.unrest.RestError as e:
            return self.idiom.data_to_response(
                dict(message=e.message, **e.extra), request, e.status
            )

        occurences = (
            f": {data['occurences']} occurences"
            if 'occurences' in data
            else ''
        )
        log.info(f'{method} {self.path}{occurences}')

        return self.idiom.data_to_response(data, request)

    def register_method(self, method):
        """
        Tells the framework to register the #route function or an overidden one
        associated with the http `method`.

        # Arguments
            method: The http method to register the route with
        """
        if method != 'OPTIONS':
            assert method in self.unrest.all, f'Unknown method {method}'

        if method not in self.methods:
            # Add method to the methods array for cohesiveness
            self.methods = (*self.methods, method)

        route = partial(self.route, method)
        route.__name__ = '_'.join((method,) + self.name_parts)
        self.unrest.framework.register_route(
            self.path, method, self.primary_keys, route
        )

        # Register options as soon as a route is registered
        if (
            self.unrest.allow_options
            and method != 'OPTIONS'
            and 'OPTIONS' not in self.methods
        ):
            self.register_method('OPTIONS')

    def has(self, pks):
        """Returns whether the pks dict has values in it."""
        return pks and all(val is not None for val in pks.values())

    def get_from_pk(self, query, **pks):
        """Get the item from `query` that has `**pks` or None if not found."""
        for key, val in pks.items():
            query = query.filter(getattr(self.Model, key) == val)
        return query.first()

    def get_all_from_pks(self, query, items_pks):
        """
        Get all items from `query` correponding to the primary keys `items_pks`
        in one query.
        """
        return query.filter(
            or_(
                *[
                    and_(
                        *[
                            getattr(self.Model, key) == val
                            for key, val in pks.items()
                        ]
                    )
                    for pks in items_pks
                ]
            )
        ).all()

    @contextmanager
    def query_request(self, request):
        """
        Context manager that sets the `_query_alterer` to the idiom alter_query
        and restore it to identity at exit.
        """
        self._query_alterer = partial(self.idiom.alter_query, request)
        yield
        self._query_alterer = _identity

    @property
    def session(self):
        """Shortcut property to the #::unrest#UnRest session."""
        return self.unrest.session

    @property
    def query(self):
        """Gets the current query associated to this Model."""
        query = getattr(self.Model, 'query', None)
        if not query or not isinstance(query, Query):
            query = self.session.query(self.Model)
        return self._query_alterer(self.query_factory(query))

    @property
    def undefered_query(self):
        """Gets the query with all attributes undefered."""
        return self.query.options(Load(self.Model).undefer('*'))

    @property
    def name_parts(self):
        """
        Returns a tuple containing optionally the schema of this table and its
        name.
        """
        if self.table.schema:
            return (self.table.schema, self.name)
        return (self.name,)

    @property
    def path(self):
        """Gets the root path of this endpoint."""
        return '/'.join((self.unrest.root_path,) + self.name_parts)

    @property
    def table(self):
        """This Model table name."""
        return self.Model.__table__

    @property
    def mapper(self):
        """Get the SQLAlchemy mapper of this Model."""
        return inspect(self.Model)

    @property
    def primary_keys(self):
        """This model primary keys names."""
        if self._primary_keys:
            return self._primary_keys

        return [
            self.mapper.get_property_by_column(pk).key
            for pk in self.mapper.primary_key
        ]

    @property
    def columns(self):
        """Gets all columns of this model `column_property` included."""

        def gen():
            for name, column in self.mapper.columns.items():
                if name not in self.primary_keys:
                    if self.only is not None and name not in self.only:
                        continue
                    if self.exclude and name in self.exclude:
                        continue
                yield name, column

        return dict(gen())



from .. import idsorted
from ...unrest import UnRest
from ..model import Fruit, Tree


def test_patch_tree(client):
    rest = UnRest(client.app, client.session, framework=client.__framework__)
    rest(Tree, methods=['GET', 'PATCH'])
    code, json = client.fetch(
        '/api/tree/1', method="PATCH", json={'id': 1, 'name': 'cedar'}
    )
    assert code == 200
    assert json['occurences'] == 1
    assert idsorted(json['objects']) == [{'id': 1, 'name': 'cedar'}]

    code, json = client.fetch('/api/tree')
    assert code == 200
    assert json['occurences'] == 3
    assert idsorted(json['objects']) == [
        {'id': 1, 'name': 'cedar'},
        {'id': 2, 'name': 'maple'},
        {'id': 3, 'name': 'oak'},
    ]


def test_patch_tree_without_id(client):
    rest = UnRest(client.app, client.session, framework=client.__framework__)
    rest(Tree, methods=['GET', 'PATCH'])
    code, json = client.fetch(
        '/api/tree/1', method="PATCH", json={'name': 'cedar'}
    )
    assert code == 200
    assert json['occurences'] == 1
    assert idsorted(json['objects']) == [{'id': 1, 'name': 'cedar'}]

    code, json = client.fetch('/api/tree')
    assert code == 200
    assert json['occurences'] == 3
    assert idsorted(json['objects']) == [
        {'id': 1, 'name': 'cedar'},
        {'id': 2, 'name': 'maple'},
        {'id': 3, 'name': 'oak'},
    ]


def test_patch_tree_missing_tree(client):
    rest = UnRest(client.app, client.session, framework=client.__framework__)
    rest(Tree, methods=['GET', 'PATCH'], allow_batch=True)
    code, json = client.fetch(
        '/api/tree/9',
        method="PATCH",
        json={
            'objects': [{'id': 1, 'name': 'cedar'}, {'id': 8, 'name': 'mango'}]
        },
    )
    assert code == 404
    assert json['message'] == "tree({'id': 9}) not found"


def test_patch_tree_with_another_id(client):
    rest = UnRest(client.app, client.session, framework=client.__framework__)
    rest(Tree, methods=['GET', 'PATCH'])
    code, json = client.fetch(
        '/api/tree/1', method="PATCH", json={'id': 5, 'name': 'cedar'}
    )
    assert code == 500


def test_patch_fruit(client):
    rest = UnRest(client.app, client.session, framework=client.__framework__)
    rest(Fruit, methods=['GET', 'PATCH'], allow_batch=True)
    code, json = client.fetch(
        '/api/fruit/1', method="PATCH", json={'fruit_id': 1, 'color': 'blue'}
    )
    assert code == 200
    assert json['occurences'] == 1
    assert idsorted(json['objects'], 'fruit_id') == [
        {
            'fruit_id': 1,
            'color': 'blue',
            'size': 12.0,
            'double_size': 24.0,
            'age': 1_041_300.0,
            'tree_id': 1,
        }
    ]

    code, json = client.fetch('/api/fruit')
    assert code == 200
    assert json['occurences'] == 5
    assert idsorted(json['objects'], 'fruit_id') == [
        {
            'fruit_id': 1,
            'color': 'blue',
            'size': 12.0,
            'double_size': 24.0,
            'age': 1_041_300.0,
            'tree_id': 1,
        },
        {
            'fruit_id': 2,
            'color': 'darkgrey',
            'size': 23.0,
            'double_size': 46.0,
            'age': 4_233_830.213,
            'tree_id': 1,
        },
        {
            'fruit_id': 3,
            'color': 'brown',
            'size': 2.12,
            'double_size': 4.24,
            'age': 0.0,
            'tree_id': 1,
        },
        {
            'fruit_id': 4,
            'color': 'red',
            'size': 0.5,
            'double_size': 1.0,
            'age': 2400.0,
            'tree_id': 2,
        },
        {
            'fruit_id': 5,
            'color': 'orangered',
            'size': 100.0,
            'double_size': 200.0,
            'age': 7200.000_012,
            'tree_id': 2,
        },
    ]



import sys

from pytest import raises

from unrest import UnRest

from .. import idsorted
from ...idiom import Idiom
from ...idiom.json_server import JsonServerIdiom
from ...idiom.unrest import UnRestIdiom
from ...idiom.yaml import YamlIdiom
from ...util import Response
from ..model import Fruit, Tree


def test_idiom(client):
    class FakeIdiom(Idiom):
        def request_to_payload(client, request):
            if request.method == 'PUT':
                return {'name': 'sth'}

        def data_to_response(client, data, method, status=200):
            payload = f"Hello {data['occurences']}"
            headers = {'Content-Type': 'text/plain'}
            response = Response(payload, headers, status)
            return response

    rest = UnRest(
        client.app,
        client.session,
        idiom=FakeIdiom,
        framework=client.__framework__,
    )
    rest(Tree, methods=['GET', 'PUT'])
    code, html = client.fetch('/api/tree')
    assert code == 200
    assert html == 'Hello 3'

    code, html = client.fetch(
        '/api/tree/1', method="PUT", json={'id': 1, 'name': 'cedar'}
    )
    assert code == 200
    assert html == 'Hello 1'

    code, html = client.fetch('/api/tree')
    assert code == 200
    assert html == 'Hello 3'


def test_idiom_alter_query(client):
    class FakeIdiom(UnRestIdiom):
        def alter_query(client, request, query):
            if request.query.get('limit'):
                query = query.limit(request.query['limit'][0])
            return query

    rest = UnRest(
        client.app,
        client.session,
        idiom=FakeIdiom,
        framework=client.__framework__,
    )
    tree = rest(Tree, methods=['GET', 'PUT'])
    assert tree.query.count() == 3

    code, json = client.fetch('/api/tree?limit=1')
    assert code == 200
    assert idsorted(json['objects']) == [{'id': 1, 'name': 'pine'}]

    code, json = client.fetch('/api/tree?limit=2')
    assert code == 200
    assert idsorted(json['objects']) == [
        {'id': 1, 'name': 'pine'},
        {'id': 2, 'name': 'maple'},
    ]

    code, json = client.fetch('/api/tree')
    assert code == 200
    assert idsorted(json['objects']) == [
        {'id': 1, 'name': 'pine'},
        {'id': 2, 'name': 'maple'},
        {'id': 3, 'name': 'oak'},
    ]

    assert tree.query.count() == 3


def test_idiom_partial_implementation(client):
    class FakeIdiom(Idiom):
        def request_to_payload(client, request):
            pass

    rest = UnRest(
        client.app,
        client.session,
        idiom=FakeIdiom,
        framework=client.__framework__,
    )
    rest(Tree, methods=['GET', 'PUT'])
    code, html = client.fetch('/api/tree')
    assert code == 500


def test_idiom_no_implementation(client):
    rest = UnRest(
        client.app, client.session, idiom=Idiom, framework=client.__framework__
    )
    rest(Tree, methods=['GET', 'PUT'])
    code, html = client.fetch('/api/tree')
    assert code == 500


def test_yaml_idiom_get(client):
    rest = UnRest(
        client.app,
        client.session,
        idiom=YamlIdiom,
        framework=client.__framework__,
    )
    rest(Tree)

    code, yaml = client.fetch('/api/tree')
    assert code == 200
    assert (
        yaml
        == '''\
objects:
- id: 1
  name: pine
- id: 2
  name: maple
- id: 3
  name: oak
occurences: 3
primary_keys:
- id
'''
    )


def test_yaml_idiom_put(client):
    rest = UnRest(
        client.app,
        client.session,
        idiom=YamlIdiom,
        framework=client.__framework__,
    )
    rest(Tree, methods=['GET', 'PUT'], allow_batch=True)

    code, yaml = client.fetch(
        '/api/tree',
        method="PUT",
        body='''\
objects:
- id: 1
  name: cedar
- id: 2
  name: mango
''',
    )
    assert (
        yaml
        == '''\
objects:
- id: 1
  name: cedar
- id: 2
  name: mango
occurences: 2
primary_keys:
- id
'''
    )

    code, yaml = client.fetch('/api/tree')
    assert code == 200
    assert (
        yaml
        == '''\
objects:
- id: 1
  name: cedar
- id: 2
  name: mango
occurences: 2
primary_keys:
- id
'''
    )


def test_yaml_idiom_put_bad_formed(client):
    rest = UnRest(
        client.app,
        client.session,
        idiom=YamlIdiom,
        framework=client.__framework__,
    )
    rest(Tree, methods=['GET', 'PUT'], allow_batch=True)

    code, yaml = client.fetch(
        '/api/tree',
        method="PUT",
        body='''\
objects: ][
''',
    )
    assert code == 400
    assert (
        yaml
        == '''message: "YAML Error in payload: while parsing a block node\
\\nexpected the node content,\\\n  \\ but found \']\'\\n  in \
\\"<unicode string>\\", line 1, column 10:\\n    objects: ][\\n\\\n  \
\\             ^"
'''
    )


def test_yaml_idiom_no_yaml(client):
    sys.modules['yaml'] = None
    rest = UnRest(
        client.app,
        client.session,
        idiom=YamlIdiom,
        framework=client.__framework__,
    )

    with raises(ImportError):
        rest(Tree, methods=['GET', 'PUT'], allow_batch=True)

    del sys.modules['yaml']


def test_yaml_idiom_empty_get_pk_as_404(client):
    rest = UnRest(
        client.app,
        client.session,
        idiom=YamlIdiom,
        framework=client.__framework__,
        empty_get_as_404=True,
    )
    rest(Tree)
    code, yaml = client.fetch('/api/tree/6')
    assert code == 404
    assert (
        yaml
        == '''\
objects: []
occurences: 0
primary_keys:
- id
'''
    )


def test_json_server_get_tree(client):
    rest = UnRest(
        client.app,
        client.session,
        idiom=JsonServerIdiom,
        framework=client.__framework__,
    )
    rest(Tree)
    rest(Fruit)
    code, json = client.fetch('/api/tree')
    assert code == 200
    assert idsorted(json) == [
        {'id': 1, 'name': 'pine'},
        {'id': 2, 'name': 'maple'},
        {'id': 3, 'name': 'oak'},
    ]


def test_json_server_post_tree(client):
    rest = UnRest(
        client.app,
        client.session,
        idiom=JsonServerIdiom,
        framework=client.__framework__,
    )
    rest(Tree, methods=['GET', 'POST'])
    code, json = client.fetch(
        '/api/tree', method="POST", json={'name': 'cedar'}
    )
    assert code == 200
    assert json == {'id': 4, 'name': 'cedar'}

    code, json = client.fetch('/api/tree')
    assert code == 200
    assert idsorted(json) == [
        {'id': 1, 'name': 'pine'},
        {'id': 2, 'name': 'maple'},
        {'id': 3, 'name': 'oak'},
        {'id': 4, 'name': 'cedar'},
    ]


def test_json_server_put_tree(client):
    rest = UnRest(
        client.app,
        client.session,
        idiom=JsonServerIdiom,
        framework=client.__framework__,
    )
    rest(Tree, methods=['GET', 'PUT'])
    code, json = client.fetch(
        '/api/tree/1', method="PUT", json={'name': 'cedar'}
    )
    assert code == 200
    assert json == {'id': 1, 'name': 'cedar'}

    code, json = client.fetch('/api/tree')
    assert code == 200
    assert idsorted(json) == [
        {'id': 1, 'name': 'cedar'},
        {'id': 2, 'name': 'maple'},
        {'id': 3, 'name': 'oak'},
    ]


def test_json_server_put_trees(client):
    rest = UnRest(
        client.app,
        client.session,
        idiom=JsonServerIdiom,
        framework=client.__framework__,
    )
    rest(Tree, methods=['GET', 'PUT'], allow_batch=True)
    code, json = client.fetch(
        '/api/tree',
        method="PUT",
        json=[{'id': 1, 'name': 'cedar'}, {'id': 2, 'name': 'mango'}],
    )
    assert code == 200
    assert idsorted(json) == [
        {'id': 1, 'name': 'cedar'},
        {'id': 2, 'name': 'mango'},
    ]

    code, json = client.fetch('/api/tree')
    assert code == 200
    assert idsorted(json) == [
        {'id': 1, 'name': 'cedar'},
        {'id': 2, 'name': 'mango'},
    ]


def test_json_server_delete_trees(client):
    rest = UnRest(
        client.app,
        client.session,
        idiom=JsonServerIdiom,
        empty_get_as_404=True,
        framework=client.__framework__,
    )
    rest(Tree, methods=['GET', 'DELETE'], allow_batch=True)
    code, json = client.fetch('/api/tree', method="DELETE")
    assert code == 200
    assert idsorted(json) == [
        {'id': 1, 'name': 'pine'},
        {'id': 2, 'name': 'maple'},
        {'id': 3, 'name': 'oak'},
    ]

    code, json = client.fetch('/api/tree')
    assert code == 404


def test_json_server_get_tree_with_relationship(client):
    rest = UnRest(
        client.app,
        client.session,
        idiom=JsonServerIdiom,
        framework=client.__framework__,
    )
    fruit = rest(Fruit, methods=[], only=['color'])

    rest(Tree, relationships={'fruits': fruit})
    code, json = client.fetch('/api/tree')
    assert code == 200
    assert idsorted(json) == [
        {'id': 1, 'name': 'pine', 'fruits': [1, 2, 3]},
        {'id': 2, 'name': 'maple', 'fruits': [4, 5]},
        {'id': 3, 'name': 'oak', 'fruits': []},
    ]
    code, json = client.fetch('/api/fruit')
    assert code == 404


def test_json_server_get_tree_with_relationship_with_multi_pk(client):
    rest = UnRest(
        client.app,
        client.session,
        idiom=JsonServerIdiom,
        framework=client.__framework__,
    )
    fruit = rest(Fruit, methods=[], primary_keys=['color', 'age'])

    rest(Tree, relationships={'fruits': fruit})
    code, json = client.fetch('/api/tree')
    assert code == 200
    assert idsorted(json) == [
        {
            'id': 1,
            'name': 'pine',
            'fruits': [
                'grey___1041300.0',
                'darkgrey___4233830.213',
                'brown___0.0',
            ],
        },
        {
            'id': 2,
            'name': 'maple',
            'fruits': ['red___2400.0', 'orangered___7200.000012'],
        },
        {'id': 3, 'name': 'oak', 'fruits': []},
    ]
    code, json = client.fetch('/api/fruit')
    assert code == 404


def test_json_server_bad_json(client):
    rest = UnRest(
        client.app,
        client.session,
        idiom=JsonServerIdiom,
        framework=client.__framework__,
    )
    rest(Tree, methods=['GET', 'POST'])
    code, json = client.fetch(
        '/api/tree', method="POST", body="{'name'; 'cedar'}"
    )
    assert code == 400
    assert (
        json['message'] == 'JSON Error in payload: '
        'Expecting property name enclosed in double quotes: '
        'line 1 column 2 (char 1)'
    )


def test_json_server_filter_fruits(client):
    rest = UnRest(
        client.app,
        client.session,
        idiom=JsonServerIdiom,
        framework=client.__framework__,
    )
    rest(Fruit)
    code, json = client.fetch('/api/fruit?color=red')
    assert code == 200
    assert idsorted(json, 'fruit_id') == [
        {
            'fruit_id': 4,
            'color': 'red',
            'size': 0.5,
            'double_size': 1.0,
            'age': 2400.0,
            'tree_id': 2,
        }
    ]
    code, json = client.fetch('/api/fruit?color=red&color=brown')
    assert code == 200
    assert idsorted(json, 'fruit_id') == [
        {
            'fruit_id': 3,
            'color': 'brown',
            'size': 2.12,
            'double_size': 4.24,
            'age': 0.0,
            'tree_id': 1,
        },
        {
            'fruit_id': 4,
            'color': 'red',
            'size': 0.5,
            'double_size': 1.0,
            'age': 2400.0,
            'tree_id': 2,
        },
    ]


def test_json_server_filter_fruits_gte(client):
    rest = UnRest(
        client.app,
        client.session,
        idiom=JsonServerIdiom,
        framework=client.__framework__,
    )
    rest(Fruit)
    code, json = client.fetch('/api/fruit?color_gte=forestgreen')
    assert code == 200
    assert idsorted(json, 'fruit_id') == [
        {
            'fruit_id': 1,
            'color': 'grey',
            'size': 12.0,
            'double_size': 24.0,
            'age': 1_041_300.0,
            'tree_id': 1,
        },
        {
            'fruit_id': 4,
            'color': 'red',
            'size': 0.5,
            'double_size': 1.0,
            'age': 2400.0,
            'tree_id': 2,
        },
        {
            'fruit_id': 5,
            'color': 'orangered',
            'size': 100.0,
            'double_size': 200.0,
            'age': 7200.000_012,
            'tree_id': 2,
        },
    ]


def test_json_server_filter_fruits_lte(client):
    rest = UnRest(
        client.app,
        client.session,
        idiom=JsonServerIdiom,
        framework=client.__framework__,
    )
    rest(Fruit)
    code, json = client.fetch('/api/fruit?size_lte=4')
    assert code == 200
    assert idsorted(json, 'fruit_id') == [
        {
            'fruit_id': 3,
            'color': 'brown',
            'size': 2.12,
            'double_size': 4.24,
            'age': 0.0,
            'tree_id': 1,
        },
        {
            'fruit_id': 4,
            'color': 'red',
            'size': 0.5,
            'double_size': 1.0,
            'age': 2400.0,
            'tree_id': 2,
        },
    ]


def test_json_server_filter_fruits_ne(client):
    rest = UnRest(
        client.app,
        client.session,
        idiom=JsonServerIdiom,
        framework=client.__framework__,
    )
    rest(Fruit)
    code, json = client.fetch('/api/fruit?age_ne=1970-01-01%2000:40:00.000000')
    assert code == 200
    assert idsorted(json, 'fruit_id') == [
        {
            'fruit_id': 1,
            'color': 'grey',
            'size': 12.0,
            'double_size': 24.0,
            'age': 1_041_300.0,
            'tree_id': 1,
        },
        {
            'fruit_id': 2,
            'color': 'darkgrey',
            'size': 23.0,
            'double_size': 46.0,
            'age': 4_233_830.213,
            'tree_id': 1,
        },
        {
            'fruit_id': 3,
            'color': 'brown',
            'size': 2.12,
            'double_size': 4.24,
            'age': 0.0,
            'tree_id': 1,
        },
        {
            'fruit_id': 5,
            'color': 'orangered',
            'size': 100.0,
            'double_size': 200.0,
            'age': 7200.000_012,
            'tree_id': 2,
        },
    ]


def test_json_server_filter_fruits_like(client):
    rest = UnRest(
        client.app,
        client.session,
        idiom=JsonServerIdiom,
        framework=client.__framework__,
    )
    rest(Fruit)
    code, json = client.fetch('/api/fruit?color_like=gre')
    assert code == 200
    assert idsorted(json, 'fruit_id') == [
        {
            'fruit_id': 1,
            'color': 'grey',
            'size': 12.0,
            'double_size': 24.0,
            'age': 1_041_300.0,
            'tree_id': 1,
        },
        {
            'fruit_id': 2,
            'color': 'darkgrey',
            'size': 23.0,
            'double_size': 46.0,
            'age': 4_233_830.213,
            'tree_id': 1,
        },
    ]


def test_json_server_filter_fruits_q(client):
    rest = UnRest(
        client.app,
        client.session,
        idiom=JsonServerIdiom,
        framework=client.__framework__,
    )
    rest(Fruit)
    code, json = client.fetch('/api/fruit?q=gre')
    assert code == 200
    assert idsorted(json, 'fruit_id') == [
        {
            'fruit_id': 1,
            'color': 'grey',
            'size': 12.0,
            'double_size': 24.0,
            'age': 1_041_300.0,
            'tree_id': 1,
        },
        {
            'fruit_id': 2,
            'color': 'darkgrey',
            'size': 23.0,
            'double_size': 46.0,
            'age': 4_233_830.213,
            'tree_id': 1,
        },
    ]


def test_json_server_sort(client):
    rest = UnRest(
        client.app,
        client.session,
        idiom=JsonServerIdiom,
        framework=client.__framework__,
    )
    rest(Tree)
    code, json = client.fetch('/api/tree?_sort=name')
    assert code == 200
    assert json == [
        {'id': 2, 'name': 'maple'},
        {'id': 3, 'name': 'oak'},
        {'id': 1, 'name': 'pine'},
    ]


def test_json_server_sort_desc(client):
    rest = UnRest(
        client.app,
        client.session,
        idiom=JsonServerIdiom,
        framework=client.__framework__,
    )
    rest(Tree)
    code, json = client.fetch('/api/tree?_sort=name&_order=desc')
    assert code == 200
    assert json == [
        {'id': 1, 'name': 'pine'},
        {'id': 3, 'name': 'oak'},
        {'id': 2, 'name': 'maple'},
    ]


def test_json_server_sort_multiple(client):
    rest = UnRest(
        client.app,
        client.session,
        idiom=JsonServerIdiom,
        framework=client.__framework__,
    )
    rest(Tree, methods=['GET', 'POST'])
    client.fetch('/api/tree', method="POST", json={'name': 'pine'})
    client.fetch('/api/tree', method="POST", json={'name': 'oak'})
    client.fetch('/api/tree', method="POST", json={'name': 'oak'})

    code, json = client.fetch('/api/tree?_sort=name,id&_order=asc,desc')
    assert code == 200
    assert json == [
        {'id': 2, 'name': 'maple'},
        {'id': 6, 'name': 'oak'},
        {'id': 5, 'name': 'oak'},
        {'id': 3, 'name': 'oak'},
        {'id': 4, 'name': 'pine'},
        {'id': 1, 'name': 'pine'},
    ]


def test_json_server_sort_rel_hack(client):
    rest = UnRest(
        client.app,
        client.session,
        idiom=JsonServerIdiom,
        framework=client.__framework__,
    )
    rest(Tree, query=lambda q: q.join(Fruit))
    code, json = client.fetch('/api/tree?_sort=fruit.size')
    assert code == 200
    assert json == [{'id': 2, 'name': 'maple'}, {'id': 1, 'name': 'pine'}]


def test_json_server_slice(client):
    rest = UnRest(
        client.app,
        client.session,
        idiom=JsonServerIdiom,
        framework=client.__framework__,
    )
    rest(Fruit, only=['color'])
    code, json = client.fetch('/api/fruit?_start=1&_end=3')
    assert code == 200
    assert json == [
        {'fruit_id': 2, 'color': 'darkgrey'},
        {'fruit_id': 3, 'color': 'brown'},
    ]


def test_json_server_paginate(client):
    rest = UnRest(
        client.app,
        client.session,
        idiom=JsonServerIdiom,
        framework=client.__framework__,
    )
    rest(Fruit, only=['color'])
    code, json = client.fetch('/api/fruit?_page=2&_limit=2')
    assert code == 200
    print(json)
    assert json == [
        {'fruit_id': 3, 'color': 'brown'},
        {'fruit_id': 4, 'color': 'red'},
    ]



from datetime import timedelta
from decimal import Decimal


class OpenApi(object):
    """
    This class generates an
    [openapi](https://github.com/OAI/OpenAPI-Specification) v3.0.0 compatible
    object from the given #::unrest.UnRest object.
    """

    __version__ = '3.0.0'

    def __init__(self, unrest):
        self.unrest = unrest

    def get_summaries_description(self, model_name):
        return {
            'GET': 'Retrieve all %s objects',
            'PUT': 'Replace all %s objects',
            'PATCH': 'Patch all %s objects',
            'POST': 'Add a new %s object',
            'DELETE': 'Delete all %s objects',
            'OPTIONS': 'Get info about the %s collection',
            'GET_one': 'Retrieve the corresponding %s object',
            'PUT_one': 'Replace the corresponding %s object or create it',
            'PATCH_one': 'Patch the corresponding %s object',
            'POST_one': 'Create a subcollection of %s object',
            'DELETE_one': 'Delete the corresponding %s object',
            'OPTIONS_one': 'Get info about the %s collection',
        }[model_name]

    def get_responses_description(self, model_name):
        return {
            'GET': 'All %s objects',
            'PUT': 'The new %s objects',
            'PATCH': 'The patched %s objects',
            'POST': 'The new %s object',
            'DELETE': 'All deleted %s objects',
            'OPTIONS': 'Info about the %s collection',
            'GET_one': 'The corresponding %s object',
            'PUT_one': 'The added %s object',
            'PATCH_one': 'The patched %s object',
            'POST_one': 'The subcollection %s object',
            'DELETE_one': 'The deleted %s object',
            'OPTIONS_one': 'Info about the %s collection',
        }[model_name]

    def get_requests_description(self, model_name):
        return {
            'PUT': 'The new %s objects to replace the current collection with',
            'PATCH': 'The %s objects patches',
            'POST': 'The new %s object to create',
            'PUT_one': 'The %s object to create or replace',
            'PATCH_one': 'The %s object patches.',
        }[model_name]

    def get_info(self):
        info = {"title": 'unrest api', "version": self.unrest.version or '1.0'}
        info.update(**self.unrest.info)
        return info

    def get_servers(self):
        return [{'url': self.unrest.framework.external_url}]

    def get_path_url(self, rest, params):
        path = rest.path
        if path.startswith(self.unrest.root_path):
            path = path[len(self.unrest.root_path) :]  # noqa
        if params:
            path += '/' + '/'.join(f'{{{pk}}}' for pk in rest.primary_keys)
        return path

    def get_property(self, type):
        try:
            type = type.python_type
        except NotImplementedError:
            type = str
        # TODO bytea b64
        if issubclass(type, bool):
            return {'type': 'boolean'}
        if issubclass(type, int):
            return {'type': 'integer', 'format': 'int64'}
        if issubclass(type, (float, Decimal)):
            return {'type': 'number', 'format': 'double'}
        if issubclass(type, timedelta):
            return {'type': 'number'}
        return {'type': 'string'}

    def get_schema(self, rest, column_only=False):
        properties = {}
        properties.update(
            {
                name: self.get_property(column.type)
                for name, column in rest.columns.items()
            }
        )
        if not column_only:
            properties.update(
                {
                    property.name: self.get_property(property.type)
                    for property in rest.properties
                }
            )
        if not column_only:
            properties.update(
                {
                    name: {
                        "type": "array",
                        "items": self.get_schema(relationship),
                    }
                    for name, relationship in rest.relationships.items()
                }
            )
        return {'type': 'object', 'properties': properties}

    def get_response_content(self, rest):
        return {
            "application/json": {
                "schema": {
                    "type": "object",
                    "properties": {
                        "primary_keys": {
                            "type": "array",
                            "items": {"type": "string"},
                        },
                        "occurences": {"type": "integer"},
                        "objects": {
                            "type": "array",
                            "items": self.get_schema(rest),
                        },
                    },
                }
            }
        }

    def get_operation_parameters(self, rest):
        return [
            {
                'name': pk,
                'in': 'path',
                'required': True,
                'schema': {'type': 'string'},
            }
            for pk in rest.primary_keys
        ]

    def get_operation_request_body(self, rest, method, params):
        model = rest.Model.__name__
        method_key = method + ('_one' if params else '')
        schema = self.get_schema(rest, True)
        if not params and method != 'POST':
            schema = {
                "type": "object",
                "properties": {"objects": {"type": "array", "items": schema}},
            }
        return {
            'description': self.get_requests_description(method_key) % model,
            'content': {"application/json": {"schema": schema}},
            'required': True,
        }

    def get_operation(self, rest, params, method):
        model = rest.Model.__name__
        method_key = method + ('_one' if params else '')
        operation = {
            "tags": ['.'.join([v for v in (rest.table.schema, model) if v])],
            "summary": self.get_summaries_description(method_key) % model,
            "responses": {
                "200": {
                    "description": self.get_responses_description(method_key)
                    % model,
                    "content": self.get_response_content(rest),
                }
            },
        }
        if params:
            operation['parameters'] = self.get_operation_parameters(rest)
        if method in ['PUT', 'PATCH', 'POST']:
            operation['requestBody'] = self.get_operation_request_body(
                rest, method, params
            )
        return operation

    def get_path(self, rest, params):
        methods = rest.methods
        if not params and not rest.allow_batch:
            methods = [
                m for m in methods if m not in ['PUT', 'DELETE', 'PATCH']
            ]
        if params:
            methods = [m for m in methods if m != 'POST']
        path = {
            method.lower(): self.get_operation(rest, params, method)
            for method in methods
        }
        if getattr(rest.Model, '__doc__', ''):
            path["summary"] = rest.Model.__doc__
        return path

    def get_paths(self):
        return {
            self.get_path_url(rest, params): self.get_path(rest, params)
            for rest in self.unrest.rests
            for params in (False, True)
        }

    def all(self):
        return {
            "openapi": self.__version__,
            "info": self.get_info(),
            "servers": self.get_servers(),
            "paths": self.get_paths(),
        }



from . import Idiom
from ..util import Response


class YamlIdiom(Idiom):
    """
    This is a very basic yaml implementation.

    It does the same thing as the #::unrest.idiom.unrest#UnRestIdiom+1 but with
    yaml instead of json.

    Requires pyyaml to be installed.
    """

    def __init__(self, rest):
        self.rest = rest
        try:
            import yaml
        except ImportError:
            raise ImportError(
                'You must have pyyaml installed to use this idiom'
            )
        self.yaml = yaml

    def request_to_payload(self, request):
        if request.payload:
            try:
                return self.yaml.load(request.payload.decode('utf-8'))
            except self.yaml.YAMLError as e:
                self.rest.raise_error(400, f'YAML Error in payload: {e}')

    def data_to_response(self, data, request, status=200):
        if (
            request.method == 'GET'
            and self.rest.unrest.empty_get_as_404
            and 'occurences' in data
            and data['occurences'] == 0
        ):
            status = 404
        payload = self.yaml.dump(data, default_flow_style=False)
        headers = {'Content-Type': 'text/yaml'}
        response = Response(payload, headers, status)
        return response



from .. import idsorted
from ...idiom.unrest import UnRestIdiom
from ...unrest import UnRest
from ..model import Tree


def test_paginated(client):
    class PaginatedUnRestIdiom(UnRestIdiom):
        def alter_query(client, request, query):
            offset = int(request.query.get('offset', [0])[0])
            step = int(request.query.get('step', [2])[0])
            return query.offset(offset).limit(step)

    rest = UnRest(
        client.app,
        client.session,
        framework=client.__framework__,
        idiom=PaginatedUnRestIdiom,
    )
    rest(Tree)
    code, json = client.fetch('/api/tree')
    assert code == 200
    assert json['occurences'] == 3
    assert json['limit'] == 2
    assert json['offset'] == 0
    assert json['primary_keys'] == ['id']
    assert idsorted(json['objects']) == [
        {'id': 1, 'name': 'pine'},
        {'id': 2, 'name': 'maple'},
    ]
    code, json = client.fetch('/api/tree?offset=1')
    assert code == 200
    assert json['occurences'] == 3
    assert json['limit'] == 2
    assert json['offset'] == 1
    assert json['primary_keys'] == ['id']
    assert idsorted(json['objects']) == [
        {'id': 2, 'name': 'maple'},
        {'id': 3, 'name': 'oak'},
    ]
    code, json = client.fetch('/api/tree?offset=2')
    assert code == 200
    assert json['occurences'] == 3
    assert json['limit'] == 2
    assert json['offset'] == 2
    assert json['primary_keys'] == ['id']
    assert idsorted(json['objects']) == [{'id': 3, 'name': 'oak'}]
    code, json = client.fetch('/api/tree?offset=3')
    assert code == 200
    assert json['occurences'] == 3
    assert json['limit'] == 2
    assert json['offset'] == 3
    assert json['primary_keys'] == ['id']
    assert idsorted(json['objects']) == []



class Request(object):
    """
    The unrest request object created in the #::unrest.framework route wrapper.

    # Arguments
        url: The request url.
        method: The request method.
        parameters: The url parameters as dict.
        query: The query string parameters as dict.
        payload: The request body as bytes if any.
        headers: A mapping of request headers.
    """

    def __init__(self, url, method, parameters, query, payload, headers):
        self.url = url
        self.method = method
        self.parameters = parameters
        self.query = query
        self.payload = payload
        self.headers = headers


class Response(object):
    """
    The unrest response object created by the #::unrest.idiom.

    # Arguments
        payload: The response body as string.
        headers: A mapping of response headers.
        status: The response status code.
    """

    def __init__(self, payload, headers, status):
        self.payload = payload
        self.headers = headers
        self.status = status



import logging
from functools import wraps

from sanic import response

from . import Framework
from ..util import Request

log = logging.getLogger(__name__)


class SanicFramework(Framework):
    """
    Unrest #::unrest.framework#Framework implementation for Sanic.
    Sadly only synchroneous sqlalchemy query are supported for now.

    Requires [Sanic](https://sanicframework.org/) to be installed.
    """

    def register_route(self, path, method, parameters, function):
        name = self._name(function.__name__)

        @wraps(function)
        def unrest_fun(request, **url_parameters):
            req = Request(
                request.url,
                request.method,
                url_parameters,
                dict(request.args),
                request.body,
                request.headers,
            )

            res = function(req)

            return response.raw(
                res.payload.encode('utf-8'),
                status=res.status,
                headers=res.headers,
            )

        self.app.add_route(unrest_fun, path, methods=[method], name=name)
        if parameters:
            params = '/'.join(f'<{param}>' for param in parameters)
            path_with_params = f'{path}/{params}'
            log.info(
                f'Registering route {name} for {path_with_params} for {method}'
            )
            self.app.add_route(
                unrest_fun, path_with_params, methods=[method], name=name
            )
        else:
            log.info(f'Registering route {name} for {path} for {method}')

    @property
    def external_url(self):
        return self.app.url_for(
            self._name('index'), _external=self.app.config.get('SERVER_NAME')
        )



from http.server import BaseHTTPRequestHandler
from io import BytesIO

from ...framework.http_server import HTTPServerFramework
from .unrest_client import UnRestClient


class FakeApp(object):
    def __init__(self, rhc):
        self.RequestHandlerClass = rhc


class FakeResponse(object):
    def __init__(self, code, headers, body):
        self.code = code
        self.headers = headers
        self.body = body


def patch_app(app):
    class FakeRequest(app.RequestHandlerClass):
        def __init__(self, request):
            self.rfile = BytesIO()
            self.rfile.write(request)
            self.rfile.seek(0)
            self.wfile = BytesIO()
            self.handle_one_request()
            self.finish()

        def log_message(self, *args):
            print(*args)

    app.RequestHandlerClass = FakeRequest


class HTTPServerClient(UnRestClient):
    __framework__ = HTTPServerFramework

    def setUp(self):
        self.get_app()
        super().setUp()

    def get_app(self):
        session = self.session

        class SimpleHTTPRequestHandler(BaseHTTPRequestHandler):
            def do_GET(self):
                if self.path != '/':
                    self.send_response(404)
                    self.end_headers()
                    return
                self.send_response(200)
                self.end_headers()
                self.wfile.write(b'A normal route!')

            def finish(self):
                session.remove()

        self.app = FakeApp(SimpleHTTPRequestHandler)
        patch_app(self.app)
        return self.app

    def raw_fetch(self, url, method='GET', headers={}, body=None):
        if body:
            headers['Content-Length'] = len(body)

        request = '\r\n'.join(
            [f'{method.upper()} {url} HTTP/1.1']
            + [f'{key}:{value}' for key, value in headers.items()]
        )
        if body:
            request += '\r\n\r\n' + body

        server = self.app.RequestHandlerClass(request.encode('iso-8859-1'))
        server.wfile.seek(0)
        res = server.wfile.read().decode('iso-8859-1')
        [head, body] = res.split('\r\n\r\n')
        res_lines = head.split('\r\n')
        [_, code, message] = res_lines[0].split(' ', 2)
        headers = {
            line.split(':')[0]: line.split(':')[1].strip()
            for line in res_lines[1:]
        }

        return FakeResponse(int(code), headers, body.encode('utf-8'))



from datetime import timedelta

from sqlalchemy.inspection import inspect

from .. import idsorted
from ...unrest import UnRest
from ..model import Fruit, Tree


def test_put_tree_validation(client):
    rest = UnRest(client.app, client.session, framework=client.__framework__)

    def name_validator(field):
        if len(field.value) > 3:
            raise rest.ValidationError(
                'Name must be shorter than 4 characters.'
            )
        return field.value

    rest(
        Tree,
        methods=['GET', 'PUT'],
        validators={'name': name_validator},
        allow_batch=True,
    )
    code, json = client.fetch(
        '/api/tree',
        method="PUT",
        json={
            'objects': [{'id': 1, 'name': 'cedar'}, {'id': 2, 'name': 'mango'}]
        },
    )
    assert code == 500
    assert json['message'] == 'Validation Error'
    assert json['errors'] == [
        {
            'id': 1,
            'fields': {'name': 'Name must be shorter than 4 characters.'},
        },
        {
            'id': 2,
            'fields': {'name': 'Name must be shorter than 4 characters.'},
        },
    ]
    code, json = client.fetch('/api/tree')
    assert code == 200
    assert json['occurences'] == 3
    assert idsorted(json['objects']) == [
        {'id': 1, 'name': 'pine'},
        {'id': 2, 'name': 'maple'},
        {'id': 3, 'name': 'oak'},
    ]
    rest(
        Tree,
        name='second_tree',
        methods=['GET', 'PUT'],
        validators={'name': name_validator},
        allow_batch=True,
    )
    code, json = client.fetch(
        '/api/tree',
        method="PUT",
        json={'objects': [{'id': 1, 'name': 'cdr'}, {'id': 2, 'name': 'mgo'}]},
    )
    assert code == 200


def test_put_pk_tree_validation(client):
    rest = UnRest(client.app, client.session, framework=client.__framework__)

    def name_validator(field):
        if len(field.value) > 3:
            raise rest.ValidationError(
                'Name must be shorter than 4 characters.'
            )
        return field.value

    rest(Tree, methods=['GET', 'PUT'], validators={'name': name_validator})
    code, json = client.fetch(
        '/api/tree/1', method="PUT", json={'id': 1, 'name': 'cedar'}
    )
    assert code == 500
    assert json['message'] == 'Validation Error'
    assert json['errors'] == [
        {
            'id': 1,
            'fields': {'name': 'Name must be shorter than 4 characters.'},
        }
    ]
    code, json = client.fetch('/api/tree')
    assert code == 200
    assert json['occurences'] == 3
    assert idsorted(json['objects']) == [
        {'id': 1, 'name': 'pine'},
        {'id': 2, 'name': 'maple'},
        {'id': 3, 'name': 'oak'},
    ]
    code, json = client.fetch(
        '/api/tree/1', method="PUT", json={'id': 1, 'name': 'cdr'}
    )
    assert code == 200


def test_post_tree_validation(client):
    rest = UnRest(client.app, client.session, framework=client.__framework__)

    def name_validator(field):
        if len(field.value) > 3:
            raise rest.ValidationError(
                'Name must be shorter than 4 characters.'
            )
        return field.value

    rest(Tree, methods=['GET', 'POST'], validators={'name': name_validator})
    code, json = client.fetch(
        '/api/tree', method="POST", json={'name': 'mango'}
    )
    assert code == 500
    assert json['message'] == 'Validation Error'
    assert json['errors'] == [
        {
            'id': None,
            'fields': {'name': 'Name must be shorter than 4 characters.'},
        }
    ]

    code, json = client.fetch('/api/tree')
    assert code == 200
    assert json['occurences'] == 3
    assert idsorted(json['objects']) == [
        {'id': 1, 'name': 'pine'},
        {'id': 2, 'name': 'maple'},
        {'id': 3, 'name': 'oak'},
    ]

    code, json = client.fetch('/api/tree', method="POST", json={'name': 'mgo'})
    assert code == 200


def test_put_fruit_dual_validation(client):
    rest = UnRest(client.app, client.session, framework=client.__framework__)

    def color_validator(field):
        if inspect(field.item).attrs.color.history.deleted[0] == 'brown':
            raise field.ValidationError(
                f'A brown fruit cannot become {field.value}'
            )
        return field.value

    def size_validator(field):
        if field.value > 48.9 and field.item.age < timedelta(weeks=2):
            raise rest.ValidationError('Fruit too big for its age')
        return field.value

    def age_validator(field):
        if field.value < timedelta(days=100):
            raise field.ValidationError('Fruit too young')
        return field.value

    def tree_id_validator(field):
        if field.value == 0:
            raise rest.ValidationError('Invalid tree_id')
        return field.value

    rest(
        Fruit,
        methods=['GET', 'PUT'],
        validators={
            'color': color_validator,
            'size': size_validator,
            'age': age_validator,
            'tree_id': tree_id_validator,
        },
    )
    code, json = client.fetch(
        '/api/fruit/3',
        method="PUT",
        json={
            'fruit_id': 3,
            'color': 'green',
            'size': 50.2,
            'age': 0,
            'tree_id': 0,
        },
    )
    assert code == 500
    assert json['message'] == 'Validation Error'
    assert json['errors'] == [
        {
            'fruit_id': 3,
            'fields': {
                'color': 'A brown fruit cannot become green',
                'size': 'Fruit too big for its age',
                'age': 'Fruit too young',
                'tree_id': 'Invalid tree_id',
            },
        }
    ]

    rest(
        Fruit,
        name='second_fruit',
        methods=['GET', 'PUT'],
        validators={
            'color': color_validator,
            'size': size_validator,
            'age': age_validator,
            'tree_id': tree_id_validator,
        },
    )
    code, json = client.fetch(
        '/api/fruit/2',
        method="PUT",
        json={
            'fruit_id': 2,
            'color': 'green',
            'size': 50.2,
            'age': 12_000_000,
            'tree_id': 3,
        },
    )
    assert code == 200


def test_validation_error_code(client):
    rest = UnRest(client.app, client.session, framework=client.__framework__)

    def name_validator(field):
        if len(field.value) > 3:
            raise rest.ValidationError(
                'Name must be shorter than 4 characters.'
            )
        return field.value

    rest(
        Tree,
        methods=['GET', 'PUT'],
        validators={'name': name_validator},
        validation_error_code=202,
    )
    code, json = client.fetch(
        '/api/tree/1', method="PUT", json={'id': 1, 'name': 'cedar'}
    )
    assert code == 202
    assert json['message'] == 'Validation Error'
    assert json['errors'] == [
        {
            'id': 1,
            'fields': {'name': 'Name must be shorter than 4 characters.'},
        }
    ]
    code, json = client.fetch('/api/tree')
    assert code == 200
    assert json['occurences'] == 3
    assert idsorted(json['objects']) == [
        {'id': 1, 'name': 'pine'},
        {'id': 2, 'name': 'maple'},
        {'id': 3, 'name': 'oak'},
    ]

    code, json = client.fetch(
        '/api/tree/1', method="PUT", json={'id': 1, 'name': 'cdr'}
    )
    assert code == 200



import logging
from functools import wraps

from tornado.web import RequestHandler, _ApplicationRouter

from . import Framework
from ..util import Request

log = logging.getLogger(__name__)


class TornadoFramework(Framework):
    """
    Unrest #::unrest.framework#Framework implementation for Tornado.

    Requires [tornado](https://www.tornadoweb.org/) to be installed.
    """

    __RequestHandlerClass__ = RequestHandler

    def __init__(self, app, url):
        super().__init__(app, url)
        self.router = _ApplicationRouter(app)
        self.app.default_router.add_rules([(url + r'(.*)', self.router)])

    def register_route(self, path, method, parameters, function):
        name = self._name(function.__name__.replace(method + '_', ''))
        # Creating an url regex that accept parameters
        if parameters:
            params = '/'.join(f'(?P<{param}>.+)' for param in parameters)
            path_with_params = f'{path}(?:/{params})?'
        else:
            path_with_params = path

        Handler = self.router.named_rules.get(path_with_params)
        if not Handler:
            Handler = type(
                name + 'Handler', (self.__RequestHandlerClass__,), {}
            )
            self.router.add_rules(
                [(path_with_params, Handler, {}, path_with_params)]
            )
        elif getattr(Handler, 'target', None):
            # If Handler has been wrapper by a Rule
            Handler = Handler.target

        if hasattr(Handler, method.lower()) and hasattr(
            getattr(Handler, method.lower()), '__wrapped__'
        ):
            raise KeyError(
                f'Method {method} is already registered for path {path}'
            )

        log.info(
            f'Registering route {name} for {path_with_params} for {method}'
        )

        @wraps(function)
        def tornado_fun(self, **url_parameters):
            request = Request(
                self.request.path,
                self.request.method,
                url_parameters,
                {
                    key: [val.decode('utf-8') for val in values]
                    for key, values in self.request.query_arguments.items()
                },
                self.request.body,
                self.request.headers,
            )

            response = function(request)
            for name, value in response.headers.items():
                self.set_header(name, value)
            self.set_status(response.status)
            self.write(response.payload)

        setattr(Handler, method.lower(), tornado_fun)

    @property
    def external_url(self):
        return self.app.reverse_url(self.url)



import json
from collections import defaultdict
from itertools import zip_longest

from sqlalchemy import asc, desc, func, or_
from sqlalchemy.sql.expression import cast
from sqlalchemy.types import String

from . import Idiom
from ..util import Response

PK_DELIM = '___'


def universal_re(attr, value, query):
    dbtype = query.session.bind.dialect.__module__.split('.')[2]
    if dbtype == 'postgresql':
        return attr.op('~')(value)  # pragma: no cover
    if dbtype == 'oracle':
        return func.REGEXP_LIKE(attr, value)  # pragma: no cover
    if dbtype in ['sqlite', 'mysql', 'sybase']:
        return attr.op('REGEXP')(value)
    # Hope that a regexp fun exists
    return func.regexp(attr, value)  # pragma: no cover


class JsonServerIdiom(Idiom):
    """
    The [JSON Server](https://github.com/typicode/json-server) idiom
    implementation.

    Currently support all basic queries and first-level (no nested props)
    filter, sort, pagination, slice, operators
    (`_gte`, `_lte`, `_ne`, `_like`)
    and `q` full-text search (which works better with
    [SQLAlchemy-Searchable](https://sqlalchemy-searchable.readthedocs.io))
    """

    def request_to_payload(self, request):
        if request.payload:
            try:
                data = json.loads(request.payload.decode('utf-8'))
            except json.JSONDecodeError as e:
                self.rest.raise_error(400, f'JSON Error in payload: {e}')
            if isinstance(data, list):
                return {'objects': data}
            return data

    def data_to_response(self, data, request, status=200):
        if (
            request.method == 'GET'
            and self.rest.unrest.empty_get_as_404
            and 'occurences' in data
            and data['occurences'] == 0
        ):
            status = 404

        if 'objects' in data:
            objects = data['objects']
            for object in objects:
                for key, relationship in self.rest.relationships.items():
                    object[key] = (
                        [
                            PK_DELIM.join(
                                str(ref[pk])
                                for pk in relationship.primary_keys
                            )
                            for ref in object[key]
                        ]
                        if len(relationship.primary_keys) > 1
                        else [
                            ref[relationship.primary_keys[0]]
                            for ref in object[key]
                        ]
                    )
            # When there's parameter it applies on a unique object
            # except from POST
            if (
                request.parameters
                and all(
                    value is not None for value in request.parameters.values()
                )
                or request.method == 'POST'
            ):
                objects = objects[0]
            payload = json.dumps(objects)
        else:
            payload = json.dumps(data)
        headers = {'Content-Type': 'application/json'}
        if 'occurences' in data:
            headers['X-Total-Count'] = data['occurences']
            headers['Access-Control-Expose-Headers'] = 'X-Total-Count'
        response = Response(payload, headers, status)
        return response

    def alter_query(self, request, query):
        Model = self.rest.Model
        params = defaultdict(str)
        filters = {}
        for param, values in request.query.items():
            if param.startswith('_'):
                params[param[1:]] = values[0]
            elif param == 'q':
                params['q'] = values[0]
            else:
                filters[param] = values

        # Filter
        if filters:
            for key, values in filters.items():
                if len(values) == 1:
                    value = values[0]
                    if key.endswith('_gte'):
                        cond = getattr(Model, key[: -len('_gte')]) >= value
                    elif key.endswith('_lte'):
                        cond = getattr(Model, key[: -len('_lte')]) <= value
                    elif key.endswith('_ne'):
                        cond = getattr(Model, key[: -len('_ne')]) != value
                    elif key.endswith('_like'):
                        cond = universal_re(
                            cast(getattr(Model, key[: -len('_like')]), String),
                            value,
                            query,
                        )
                    else:
                        cond = getattr(Model, key) == value
                    query = query.filter(cond)
                else:
                    col = getattr(Model, key)
                    query = query.filter(col.in_(values))

        # Search
        if params['q']:
            if hasattr(query, 'search'):
                query = query.search(params['q'])  # pragma: no cover
            else:
                query = query.filter(
                    or_(
                        *[
                            universal_re(
                                cast(getattr(Model, column), String),
                                params['q'],
                                query,
                            )
                            for column in self.rest.columns
                        ]
                    )
                )

        # Order
        if params['sort'] and request.method == 'GET':
            for sort, way in zip_longest(
                params['sort'].split(','),
                params['order'].split(','),
                fillvalue='asc',
            ):
                way = desc if way.lower() == 'desc' else asc

                if hasattr(Model, sort):
                    query = query.order_by(way(getattr(Model, sort)))
                else:
                    query = query.order_by(way(sort.split('.')[-1]))

        if request.method == 'GET':
            query = query.order_by(
                *[getattr(Model, pk) for pk in self.rest.primary_keys]
            )

        # Offset / Limit
        offset = int(params['start'] or 0)

        if params['page']:
            step = int(params['limit'] or 10)
            offset = max(int(params['page']) - 1, 0) * step

        elif params['end']:
            step = int(params['end']) - offset
        else:
            step = int(params['limit']) if params['limit'] else None

        if step:
            query = query.offset(offset).limit(step)

        return query



import json

from . import Idiom
from ..util import Response


class UnRestIdiom(Idiom):
    """
    The default UnRest implementation.

    Parses request payload as json.
    Serialize data as json.
    Can return a 404 on empty GET if `empty_get_as_404` is set as True in the
    Unrest instance.
    """

    def request_to_payload(self, request):
        if request.payload:
            try:
                return json.loads(request.payload.decode('utf-8'))
            except json.JSONDecodeError as e:
                self.rest.raise_error(400, f'JSON Error in payload: {e}')

    def data_to_response(self, data, request, status=200):
        if (
            request.method == 'GET'
            and self.rest.unrest.empty_get_as_404
            and 'occurences' in data
            and data['occurences'] == 0
        ):
            status = 404
        payload = json.dumps(data)
        headers = {'Content-Type': 'application/json'}
        response = Response(payload, headers, status)
        return response



from datetime import timedelta

from .. import idsorted
from ...unrest import UnRest
from ..model import Fruit, Tree


def test_patch_tree_implicitly_unallowed(client):
    rest = UnRest(client.app, client.session, framework=client.__framework__)
    rest(Tree, methods=['GET', 'PATCH'])
    code, json = client.fetch(
        '/api/tree', method="PATCH", json={'id': 1, 'name': 'cedar'}
    )
    assert code == 406


def test_patch_empty_payload(client):
    rest = UnRest(client.app, client.session, framework=client.__framework__)
    rest(Tree, methods=['GET', 'PATCH'])
    code, json = client.fetch('/api/tree', method="PATCH", json={})
    assert code == 400
    assert json['message'] == 'You must provide a payload'


def test_patch_tree(client):
    rest = UnRest(client.app, client.session, framework=client.__framework__)
    rest(Tree, methods=['GET', 'PATCH'], allow_batch=True)
    code, json = client.fetch(
        '/api/tree',
        method="PATCH",
        json={
            'objects': [{'id': 1, 'name': 'cedar'}, {'id': 2, 'name': 'mango'}]
        },
    )
    assert code == 200
    assert json['occurences'] == 2
    assert idsorted(json['objects']) == [
        {'id': 1, 'name': 'cedar'},
        {'id': 2, 'name': 'mango'},
    ]

    code, json = client.fetch('/api/tree')
    assert code == 200
    assert json['occurences'] == 3
    assert idsorted(json['objects']) == [
        {'id': 1, 'name': 'cedar'},
        {'id': 2, 'name': 'mango'},
        {'id': 3, 'name': 'oak'},
    ]


def test_patch_missing_tree(client):
    rest = UnRest(client.app, client.session, framework=client.__framework__)
    rest(Tree, methods=['GET', 'PATCH'], allow_batch=True)
    code, json = client.fetch(
        '/api/tree',
        method="PATCH",
        json={
            'objects': [{'id': 1, 'name': 'cedar'}, {'id': 8, 'name': 'mango'}]
        },
    )
    assert code == 404


def test_patch_fruit(client):
    rest = UnRest(client.app, client.session, framework=client.__framework__)
    rest(Fruit, methods=['GET', 'PATCH'], allow_batch=True)
    code, json = client.fetch(
        '/api/fruit',
        method="PATCH",
        json={
            'objects': [
                {'fruit_id': 1, 'color': 'blue'},
                {
                    'fruit_id': 3,
                    'age': timedelta(days=12, minutes=29).total_seconds(),
                },
                {'fruit_id': 4, 'color': 'rainbow', 'size': 8},
                {'fruit_id': 5, 'size': 10, 'tree_id': 1},
            ]
        },
    )
    assert code == 200
    assert json['occurences'] == 4
    assert idsorted(json['objects'], 'fruit_id') == [
        {
            'fruit_id': 1,
            'color': 'blue',
            'size': 12.0,
            'double_size': 24.0,
            'age': 1_041_300.0,
            'tree_id': 1,
        },
        {
            'fruit_id': 3,
            'color': 'brown',
            'size': 2.12,
            'double_size': 4.24,
            'age': 1_038_540.0,
            'tree_id': 1,
        },
        {
            'fruit_id': 4,
            'color': 'rainbow',
            'size': 8.0,
            'double_size': 16.0,
            'age': 2400.0,
            'tree_id': 2,
        },
        {
            'fruit_id': 5,
            'color': 'orangered',
            'size': 10.0,
            'double_size': 20.0,
            'age': 7200.000_012,
            'tree_id': 1,
        },
    ]

    code, json = client.fetch('/api/fruit')
    assert code == 200
    assert json['occurences'] == 5
    assert idsorted(json['objects'], 'fruit_id') == [
        {
            'fruit_id': 1,
            'color': 'blue',
            'size': 12.0,
            'double_size': 24.0,
            'age': 1_041_300.0,
            'tree_id': 1,
        },
        {
            'fruit_id': 2,
            'color': 'darkgrey',
            'size': 23.0,
            'double_size': 46.0,
            'age': 4_233_830.213,
            'tree_id': 1,
        },
        {
            'fruit_id': 3,
            'color': 'brown',
            'size': 2.12,
            'double_size': 4.24,
            'age': 1_038_540.0,
            'tree_id': 1,
        },
        {
            'fruit_id': 4,
            'color': 'rainbow',
            'size': 8.0,
            'double_size': 16.0,
            'age': 2400.0,
            'tree_id': 2,
        },
        {
            'fruit_id': 5,
            'color': 'orangered',
            'size': 10.0,
            'double_size': 20.0,
            'age': 7200.000_012,
            'tree_id': 1,
        },
    ]



from .. import idsorted
from ...unrest import UnRest
from ..model import Fruit, Tree


def test_get_pk_tree(client):
    rest = UnRest(client.app, client.session, framework=client.__framework__)
    rest(Tree)
    rest(Fruit)
    code, json = client.fetch('/api/tree/1')
    assert code == 200
    assert json['occurences'] == 1
    assert idsorted(json['objects']) == [{'id': 1, 'name': 'pine'}]


def test_get_pk_unknown_tree(client):
    rest = UnRest(client.app, client.session, framework=client.__framework__)
    rest(Tree)
    rest(Fruit)
    code, json = client.fetch('/api/tree/6')
    assert code == 200
    assert json['occurences'] == 0
    assert json['objects'] == []


def test_get_pk_tree_name(client):
    rest = UnRest(client.app, client.session, framework=client.__framework__)
    rest(Tree, name='forest')
    code, json = client.fetch('/api/forest/2')
    assert code == 200
    assert json['occurences'] == 1
    assert idsorted(json['objects']) == [{'id': 2, 'name': 'maple'}]


def test_get_pk_tree_query(client):
    rest = UnRest(client.app, client.session, framework=client.__framework__)

    def base_query(q):
        return client.session.query(Tree).filter(Tree.id > 1)

    rest(Tree, query=base_query)
    code, json = client.fetch('/api/tree/2')
    assert code == 200
    assert json['occurences'] == 1
    assert idsorted(json['objects']) == [{'id': 2, 'name': 'maple'}]


def test_get_pk_tree_query_not_found(client):
    rest = UnRest(client.app, client.session, framework=client.__framework__)

    def base_query(q):
        return client.session.query(Tree).filter(Tree.id > 1)

    rest(Tree, query=base_query)
    code, json = client.fetch('/api/tree/1')
    assert code == 200
    assert json['occurences'] == 0
    assert json['objects'] == []


def test_get_pk_tree_query_factory(client):
    rest = UnRest(client.app, client.session, framework=client.__framework__)
    rest(Tree, query=lambda q: q.filter(Tree.id < 2))
    code, json = client.fetch('/api/tree')
    assert code == 200
    assert json['occurences'] == 1
    assert idsorted(json['objects']) == [{'id': 1, 'name': 'pine'}]


def test_get_pk_tree_query_factory_not_found(client):
    rest = UnRest(client.app, client.session, framework=client.__framework__)
    rest(Tree, query=lambda q: q.filter(Tree.id < 2))
    code, json = client.fetch('/api/tree/3')
    assert code == 200
    assert json['occurences'] == 0
    assert json['objects'] == []


def test_get_pk_fruits(client):
    rest = UnRest(client.app, client.session, framework=client.__framework__)
    rest(Tree)
    rest(Fruit)
    code, json = client.fetch('/api/fruit/1')
    assert code == 200
    assert json['occurences'] == 1
    assert idsorted(json['objects'], 'fruit_id') == [
        {
            'fruit_id': 1,
            'color': 'grey',
            'size': 12.0,
            'double_size': 24.0,
            'age': 1_041_300.0,
            'tree_id': 1,
        }
    ]


def test_get_pk_fruits_only(client):
    rest = UnRest(client.app, client.session, framework=client.__framework__)
    rest(Fruit, only=['color', 'size'])
    code, json = client.fetch('/api/fruit/2')
    assert code == 200
    assert json['occurences'] == 1
    assert idsorted(json['objects'], 'fruit_id') == [
        {'fruit_id': 2, 'color': 'darkgrey', 'size': 23.0}
    ]


def test_get_pk_fruits_exclude(client):
    rest = UnRest(client.app, client.session, framework=client.__framework__)
    rest(Fruit, exclude=['color', 'age', 'double_size'])
    code, json = client.fetch('/api/fruit/3')
    assert code == 200
    assert json['occurences'] == 1
    assert idsorted(json['objects'], 'fruit_id') == [
        {'fruit_id': 3, 'size': 2.12, 'tree_id': 1}
    ]


def test_get_pk_fruits_only_exclude(client):
    rest = UnRest(client.app, client.session, framework=client.__framework__)
    rest(Fruit, only=['color', 'size'], exclude=['color', 'age'])
    code, json = client.fetch('/api/fruit/4')
    assert code == 200
    assert json['occurences'] == 1
    assert idsorted(json['objects'], 'fruit_id') == [
        {'fruit_id': 4, 'size': 0.5}
    ]


def test_no_method(client):
    rest = UnRest(client.app, client.session, framework=client.__framework__)
    rest(Fruit, methods=[])
    code, html = client.fetch('/api/fruit/1')
    assert code == 404


def test_get_custom(client):
    rest = UnRest(client.app, client.session, framework=client.__framework__)
    fruit = rest(Fruit)

    @fruit.declare('GET')
    def get(payload, fruit_id=None):
        return {'Hey': 'Overridden'}

    assert get is not None

    code, json = client.fetch('/api/fruit/3')
    assert code == 200
    assert json == {'Hey': 'Overridden'}


def test_get_custom_extend(client):
    rest = UnRest(client.app, client.session, framework=client.__framework__)
    fruit = rest(Fruit)

    @fruit.declare('GET')
    def get(payload, fruit_id=None):
        fruit_id += 1
        return fruit.get(payload, fruit_id=fruit_id)

    code, json = client.fetch('/api/fruit/4')
    assert code == 200
    assert json['occurences'] == 1
    assert idsorted(json['objects'], 'fruit_id') == [
        {
            'fruit_id': 5,
            'color': 'orangered',
            'size': 100.0,
            'double_size': 200.0,
            'age': 7200.000_012,
            'tree_id': 2,
        }
    ]


def test_declare_without_method(client):
    rest = UnRest(client.app, client.session, framework=client.__framework__)
    tree = rest(Tree)

    @tree.declare('POST')
    def post(payload, id=None):
        return {'Hey': 'Overridden'}

    assert post is not None

    code, json = client.fetch(
        '/api/tree', method="POST", json={'name': 'cedar'}
    )
    assert code == 200
    assert json == {'Hey': 'Overridden'}



from tornado.testing import AsyncHTTPTestCase
from tornado.web import Application, RequestHandler

from ...framework.tornado import TornadoFramework
from .unrest_client import UnRestClient


class SessionRemoverRequestHandler(RequestHandler):
    def on_finish(self):
        super().on_finish()
        self.application.session.remove()


class SessionRemoverTornadoFramework(TornadoFramework):
    __RequestHandlerClass__ = SessionRemoverRequestHandler


class TornadoAsyncClient(AsyncHTTPTestCase):
    def __init__(self, get_app_fun):
        super().__init__()
        self.get_app_fun = get_app_fun

    def get_app(self):
        return self.get_app_fun()

    def runTest(self):
        pass  # pragma: no cover


class TornadoClient(UnRestClient):
    __framework__ = SessionRemoverTornadoFramework

    def setUp(self):
        self.http_client = TornadoAsyncClient(self.get_app)
        self.http_client.setUp()
        super().setUp()

    def get_app(self):
        class MainHandler(SessionRemoverRequestHandler):
            def get(self):
                self.write("A normal route!")

        self.app = Application([(r"/", MainHandler)])
        self.app.session = self.session
        return self.app

    def raw_fetch(self, *args, **kwargs):
        return self.http_client.fetch(*args, **kwargs)

    def tearDown(self):
        self.http_client.tearDown()



from .. import idsorted
from ...unrest import UnRest
from ..model import Fruit, Tree


def test_get_tree(client):
    rest = UnRest(client.app, client.session, framework=client.__framework__)
    rest(Tree)
    rest(Fruit)
    code, json = client.fetch('/api/tree')
    assert code == 200
    assert json['occurences'] == 3
    assert idsorted(json['objects']) == [
        {'id': 1, 'name': 'pine'},
        {'id': 2, 'name': 'maple'},
        {'id': 3, 'name': 'oak'},
    ]


def test_get_tree_name(client):
    rest = UnRest(client.app, client.session, framework=client.__framework__)
    rest(Tree, name='forest')
    code, json = client.fetch('/api/forest')
    assert code == 200
    assert json['occurences'] == 3
    assert idsorted(json['objects']) == [
        {'id': 1, 'name': 'pine'},
        {'id': 2, 'name': 'maple'},
        {'id': 3, 'name': 'oak'},
    ]


def test_get_tree_query(client):
    rest = UnRest(client.app, client.session, framework=client.__framework__)

    def base_query(q):
        return client.session.query(Tree).filter(Tree.id > 1)

    rest(Tree, query=base_query)
    code, json = client.fetch('/api/tree')
    assert code == 200
    assert json['occurences'] == 2
    assert idsorted(json['objects']) == [
        {'id': 2, 'name': 'maple'},
        {'id': 3, 'name': 'oak'},
    ]


def test_get_tree_query_factory(client):
    rest = UnRest(client.app, client.session, framework=client.__framework__)
    rest(Tree, query=lambda q: q.filter(Tree.id < 2))
    code, json = client.fetch('/api/tree')
    assert code == 200
    assert json['occurences'] == 1
    assert json['primary_keys'] == ['id']
    assert idsorted(json['objects']) == [{'id': 1, 'name': 'pine'}]


def test_get_fruits(client):
    rest = UnRest(client.app, client.session, framework=client.__framework__)
    rest(Tree)
    rest(Fruit)
    code, json = client.fetch('/api/fruit')
    assert code == 200
    assert json['occurences'] == 5
    assert json['primary_keys'] == ['fruit_id']
    assert idsorted(json['objects'], 'fruit_id') == [
        {
            'fruit_id': 1,
            'color': 'grey',
            'size': 12.0,
            'double_size': 24.0,
            'age': 1_041_300.0,
            'tree_id': 1,
        },
        {
            'fruit_id': 2,
            'color': 'darkgrey',
            'size': 23.0,
            'double_size': 46.0,
            'age': 4_233_830.213,
            'tree_id': 1,
        },
        {
            'fruit_id': 3,
            'color': 'brown',
            'size': 2.12,
            'double_size': 4.24,
            'age': 0.0,
            'tree_id': 1,
        },
        {
            'fruit_id': 4,
            'color': 'red',
            'size': 0.5,
            'double_size': 1.0,
            'age': 2400.0,
            'tree_id': 2,
        },
        {
            'fruit_id': 5,
            'color': 'orangered',
            'size': 100.0,
            'double_size': 200.0,
            'age': 7200.000_012,
            'tree_id': 2,
        },
    ]


def test_get_fruits_only(client):
    rest = UnRest(client.app, client.session, framework=client.__framework__)
    rest(Fruit, only=['color', 'size'])
    code, json = client.fetch('/api/fruit')
    assert code == 200
    assert json['occurences'] == 5
    assert idsorted(json['objects'], 'fruit_id') == [
        {'fruit_id': 1, 'color': 'grey', 'size': 12.0},
        {'fruit_id': 2, 'color': 'darkgrey', 'size': 23.0},
        {'fruit_id': 3, 'color': 'brown', 'size': 2.12},
        {'fruit_id': 4, 'color': 'red', 'size': 0.5},
        {'fruit_id': 5, 'color': 'orangered', 'size': 100.0},
    ]


def test_get_fruits_exclude(client):
    rest = UnRest(client.app, client.session, framework=client.__framework__)
    rest(Fruit, exclude=['color', 'age', 'double_size'])
    code, json = client.fetch('/api/fruit')
    assert code == 200
    assert json['occurences'] == 5
    assert idsorted(json['objects'], 'fruit_id') == [
        {'fruit_id': 1, 'size': 12.0, 'tree_id': 1},
        {'fruit_id': 2, 'size': 23.0, 'tree_id': 1},
        {'fruit_id': 3, 'size': 2.12, 'tree_id': 1},
        {'fruit_id': 4, 'size': 0.5, 'tree_id': 2},
        {'fruit_id': 5, 'size': 100.0, 'tree_id': 2},
    ]


def test_get_fruits_only_exclude(client):
    rest = UnRest(client.app, client.session, framework=client.__framework__)
    rest(Fruit, only=['color', 'size'], exclude=['color', 'age'])
    code, json = client.fetch('/api/fruit')
    assert code == 200
    assert json['occurences'] == 5
    assert idsorted(json['objects'], 'fruit_id') == [
        {'fruit_id': 1, 'size': 12.0},
        {'fruit_id': 2, 'size': 23.0},
        {'fruit_id': 3, 'size': 2.12},
        {'fruit_id': 4, 'size': 0.5},
        {'fruit_id': 5, 'size': 100.0},
    ]


def test_no_method(client):
    rest = UnRest(client.app, client.session, framework=client.__framework__)
    rest(Fruit, methods=[])
    code, html = client.fetch('/api/fruit')
    assert code == 404


def test_get_custom(client):
    rest = UnRest(client.app, client.session, framework=client.__framework__)
    fruit = rest(Fruit)

    @fruit.declare('GET')
    def get(payload, fruit_id=None):
        return {'Hey': 'Overridden'}

    code, json = client.fetch('/api/fruit')
    assert code == 200
    assert json == {'Hey': 'Overridden'}


def test_get_custom_extend(client):
    rest = UnRest(client.app, client.session, framework=client.__framework__)
    fruit = rest(Fruit)

    @fruit.declare('GET')
    def get(payload, fruit_id=None):
        rv = fruit.get(payload, fruit_id=fruit_id)
        return {
            'occurences': rv['occurences'],
            'objects': [{'id': obj['fruit_id']} for obj in rv['objects']],
        }

    code, json = client.fetch('/api/fruit')
    assert code == 200
    assert json['occurences'] == 5
    assert idsorted(json['objects']) == [
        {'id': 1},
        {'id': 2},
        {'id': 3},
        {'id': 4},
        {'id': 5},
    ]



from .. import idsorted
from ...unrest import UnRest
from ..model import Fruit, Tree


def test_post_tree(client):
    rest = UnRest(client.app, client.session, framework=client.__framework__)
    rest(Tree, methods=['GET', 'POST'])
    code, json = client.fetch(
        '/api/tree', method="POST", json={'name': 'cedar'}
    )
    assert code == 200
    assert json['occurences'] == 1
    assert idsorted(json['objects']) == [{'id': 4, 'name': 'cedar'}]

    code, json = client.fetch('/api/tree')
    assert code == 200
    assert json['occurences'] == 4
    assert idsorted(json['objects']) == [
        {'id': 1, 'name': 'pine'},
        {'id': 2, 'name': 'maple'},
        {'id': 3, 'name': 'oak'},
        {'id': 4, 'name': 'cedar'},
    ]


def test_post_tree_null_value(client):
    rest = UnRest(client.app, client.session, framework=client.__framework__)
    rest(Tree, methods=['GET', 'POST'])
    code, json = client.fetch('/api/tree', method="POST", json={'name': None})
    assert code == 200
    assert json['occurences'] == 1
    assert idsorted(json['objects']) == [{'id': 4, 'name': None}]

    code, json = client.fetch('/api/tree')
    assert code == 200
    assert json['occurences'] == 4
    assert idsorted(json['objects']) == [
        {'id': 1, 'name': 'pine'},
        {'id': 2, 'name': 'maple'},
        {'id': 3, 'name': 'oak'},
        {'id': 4, 'name': None},
    ]


def test_post_empty_payload(client):
    rest = UnRest(client.app, client.session, framework=client.__framework__)
    rest(Tree, methods=['GET', 'POST'])
    code, json = client.fetch('/api/tree', method="POST", json={})
    assert code == 400
    assert json['message'] == 'You must provide a payload'


def test_post_tree_with_id(client):
    rest = UnRest(client.app, client.session, framework=client.__framework__)
    rest(Tree, methods=['GET', 'POST'])
    code, json = client.fetch(
        '/api/tree', method="POST", json={'id': 9, 'name': 'cedar'}
    )
    assert code == 200
    assert json['occurences'] == 1
    assert idsorted(json['objects']) == [{'id': 9, 'name': 'cedar'}]

    code, json = client.fetch('/api/tree')
    assert code == 200
    assert json['occurences'] == 4
    assert idsorted(json['objects']) == [
        {'id': 1, 'name': 'pine'},
        {'id': 2, 'name': 'maple'},
        {'id': 3, 'name': 'oak'},
        {'id': 9, 'name': 'cedar'},
    ]


def test_post_tree_custom(client):
    rest = UnRest(client.app, client.session, framework=client.__framework__)
    tree = rest(Tree)

    @tree.declare('POST')
    def post(payload, id=None):
        payload['name'] = 'I ALWAYS WANT THIS NAME'
        return tree.post(payload, id=id)

    code, json = client.fetch(
        '/api/tree', method="POST", json={'name': 'cedar'}
    )
    assert code == 200
    assert json['occurences'] == 1
    assert idsorted(json['objects']) == [
        {'id': 4, 'name': 'I ALWAYS WANT THIS NAME'}
    ]

    code, json = client.fetch('/api/tree')
    assert code == 200
    assert json['occurences'] == 4
    assert idsorted(json['objects']) == [
        {'id': 1, 'name': 'pine'},
        {'id': 2, 'name': 'maple'},
        {'id': 3, 'name': 'oak'},
        {'id': 4, 'name': 'I ALWAYS WANT THIS NAME'},
    ]


def test_post_tree_custom_manual_commit(client):
    rest = UnRest(client.app, client.session, framework=client.__framework__)
    tree = rest(Tree)

    @tree.declare('POST', True)
    def post(payload, id=None):
        payload['name'] = 'I ALWAYS WANT THIS NAME'
        return tree.post(payload, id=id)

    code, json = client.fetch(
        '/api/tree', method="POST", json={'name': 'cedar'}
    )
    assert code == 200
    assert json['occurences'] == 1
    assert idsorted(json['objects']) == [
        {'id': 4, 'name': 'I ALWAYS WANT THIS NAME'}
    ]
    # The post should not be commited
    code, json = client.fetch('/api/tree')
    assert code == 200
    assert json['occurences'] == 3
    assert idsorted(json['objects']) == [
        {'id': 1, 'name': 'pine'},
        {'id': 2, 'name': 'maple'},
        {'id': 3, 'name': 'oak'},
    ]


def test_post_with_defaults(client):
    rest = UnRest(client.app, client.session, framework=client.__framework__)
    rest(
        Fruit,
        methods=['GET', 'POST'],
        defaults={'color': 'white', 'age': lambda p: p['size'] * 2},
    )
    code, json = client.fetch(
        '/api/fruit', method="POST", json={'size': 1.0, 'tree_id': 1}
    )
    assert code == 200
    assert json['occurences'] == 1
    assert idsorted(json['objects'], 'fruit_id') == [
        {
            'fruit_id': 6,
            'color': 'white',
            'size': 1.0,
            'double_size': 2.0,
            'age': 2.0,
            'tree_id': 1,
        }
    ]
    code, json = client.fetch(
        '/api/fruit',
        method="POST",
        json={'color': 'yellow', 'size': 2.0, 'tree_id': 2},
    )
    assert code == 200
    assert json['occurences'] == 1
    assert idsorted(json['objects'], 'fruit_id') == [
        {
            'fruit_id': 7,
            'color': 'yellow',
            'size': 2.0,
            'double_size': 4.0,
            'age': 4.0,
            'tree_id': 2,
        }
    ]

    code, json = client.fetch('/api/fruit')
    assert code == 200
    assert json['occurences'] == 7
    assert idsorted(json['objects'][-2:], 'fruit_id') == [
        {
            'fruit_id': 6,
            'color': 'white',
            'size': 1.0,
            'double_size': 2.0,
            'age': 2.0,
            'tree_id': 1,
        },
        {
            'fruit_id': 7,
            'color': 'yellow',
            'size': 2.0,
            'double_size': 4.0,
            'age': 4.0,
            'tree_id': 2,
        },
    ]


def test_post_with_fixed(client):
    rest = UnRest(client.app, client.session, framework=client.__framework__)
    rest(
        Fruit,
        methods=['GET', 'POST'],
        fixed={'color': 'white', 'age': lambda p: p['size'] * 2},
    )
    code, json = client.fetch(
        '/api/fruit', method="POST", json={'size': 1.0, 'tree_id': 1}
    )
    assert code == 200
    assert json['occurences'] == 1
    assert idsorted(json['objects'], 'fruit_id') == [
        {
            'fruit_id': 6,
            'color': 'white',
            'size': 1.0,
            'double_size': 2.0,
            'age': 2.0,
            'tree_id': 1,
        }
    ]
    code, json = client.fetch(
        '/api/fruit',
        method="POST",
        json={'color': 'yellow', 'size': 2.0, 'tree_id': 2},
    )
    assert code == 200
    assert json['occurences'] == 1
    assert idsorted(json['objects'], 'fruit_id') == [
        {
            'fruit_id': 7,
            'color': 'white',
            'size': 2.0,
            'double_size': 4.0,
            'age': 4.0,
            'tree_id': 2,
        }
    ]

    code, json = client.fetch('/api/fruit')
    assert code == 200
    assert json['occurences'] == 7
    assert idsorted(json['objects'][-2:], 'fruit_id') == [
        {
            'fruit_id': 6,
            'color': 'white',
            'size': 1.0,
            'double_size': 2.0,
            'age': 2.0,
            'tree_id': 1,
        },
        {
            'fruit_id': 7,
            'color': 'white',
            'size': 2.0,
            'double_size': 4.0,
            'age': 4.0,
            'tree_id': 2,
        },
    ]



import logging
import re
from types import MethodType
from urllib.parse import parse_qs, urlparse

from . import Framework
from ..util import Request

log = logging.getLogger(__name__)


class HTTPServerFramework(Framework):
    """
    Unrest #::unrest.framework#Framework implementation for
    [http.server.HTTPServer](https://docs.python.org/3/library/http.server.html)
    compatible app.

    This exemple implementation requires no external library.
    """

    def __init__(self, app, url):
        super().__init__(app, url)
        self.url_map = {}
        parent = self

        class HTTPServerFrameworkHandlerClass(self.app.RequestHandlerClass):
            """
            http.server.RequestHandlerClass implementation for UnRest
            #Framework
            """

            def __getattribute__(self, name):
                if name.startswith('do_'):
                    method = name.replace('do_', '')

                    def do_METHOD(self):
                        # Handle only requests starting with url
                        if not self.path.startswith(parent.url):
                            return getattr(super(), name)()
                        return self.handle_request(method)

                    return MethodType(do_METHOD, self)
                return super().__getattribute__(name)

            def handle_request(self, method):
                url = urlparse(self.path)
                # Look up in the url_map if we have matching endpoint
                for path, methods in parent.url_map.items():
                    match = re.fullmatch(path, url.path)
                    if match:
                        # With a corresponding method
                        if method not in methods:
                            return self.send(405, 'Method Not Allowed')
                        return self.respond(
                            url, method, methods[method], match.groupdict()
                        )
                return self.send(404, 'Not Found')

            def send(self, status, message, headers=None):
                headers = headers or {}
                self.send_response(status)

                for name, value in headers.items():
                    self.send_header(name, value)
                self.end_headers()

                self.wfile.write(message.encode('utf-8'))

            def respond(self, url, method, function, url_parameters):
                length = (
                    int(self.headers['Content-Length'])
                    if 'Content-Length' in self.headers
                    else 0
                )
                body = self.rfile.read(length) if length else ''
                request = Request(
                    url.path,
                    method,
                    url_parameters,
                    parse_qs(url.query),
                    body,
                    self.headers,
                )
                try:
                    response = function(request)
                except Exception:
                    log.exception(f'Error on {method} {self.path}')
                    return self.send(500, 'Internal Server Error')

                self.send(response.status, response.payload, response.headers)

        self.app.RequestHandlerClass = HTTPServerFrameworkHandlerClass

    def register_route(self, path, method, parameters, function):
        name = self._name(function.__name__.replace(method + '_', ''))
        # Creating an url regex that accept parameters
        if parameters:
            params = '/'.join(f'(?P<{param}>.+)' for param in parameters)
            path_with_params = f'{path}(?:/{params})?'
        else:
            path_with_params = path

        # If this is the first method for path, initialize method mapping
        if path_with_params not in self.url_map:
            self.url_map[path_with_params] = {}

        if method in self.url_map[path_with_params]:
            raise KeyError(
                f'Method {method} is already registered for path {path}'
            )

        log.info(
            f'Registering route {name} for {path_with_params} for {method}'
        )

        # Associate UnRest function with path and method
        self.url_map[path_with_params][method] = function



import logging
from functools import wraps

from flask import request as flask_request
from flask import url_for

from . import Framework
from ..util import Request

log = logging.getLogger(__name__)


class FlaskFramework(Framework):
    """
    Unrest #::unrest.framework#Framework implementation for Flask.

    Requires [flask](http://flask.pocoo.org/) to be installed.
    """

    def register_route(self, path, method, parameters, function):
        name = self._name(function.__name__)
        getattr(
            function, '__func__', function
        ).provide_automatic_options = False

        @wraps(function)
        def unrest_fun(**url_parameters):
            request = Request(
                flask_request.url,
                flask_request.method,
                url_parameters,
                dict(flask_request.args.lists()),
                flask_request.data,
                flask_request.headers,
            )

            response = function(request)

            return self.app.response_class(
                response.payload,
                status=response.status,
                headers=response.headers,
            )

        self.app.add_url_rule(path, name, unrest_fun, methods=[method])
        if parameters:
            params = '/'.join(f'<{param}>' for param in parameters)
            path_with_params = f'{path}/{params}'

            log.info(
                f'Registering route {name} for {path_with_params} for {method}'
            )
            self.app.add_url_rule(
                path_with_params, name, unrest_fun, methods=[method]
            )
        else:
            log.info(f'Registering route {name} for {path} for {method}')

    @property
    def external_url(self):
        return url_for(self._name('index'), _external=True)



class Idiom(object):
    """
    UnRest Idiom abstract class.

    Idioms are a way to alter how UnRest speaks with the rest of the world.
    The default implementation #::unrest.idiom.unrest#UnRestIdiom+1 parses json
    from request and generates json from response.

    To implement an idiom you have to implement the `request_to_payload` and
    the `data_to_response` methods described below.

    You can also override the `alter_query` method to add functionality to the
    idiom based on the `request`.
    (i.e. implementing sort / filter query parameters)

    # Arguments
        rest: The current rest instance
    """

    def __init__(self, rest):
        self.rest = rest

    def request_to_payload(self, request):
        """
        This method takes a #::unrest.util#Request `request` parameter and
        returns a payload dict.

        # Arguments
            request: The #::unrest.util#Request request

        # Returns
        A dict containing the request payload data.
        """
        raise NotImplementedError()

    def data_to_response(self, data, request, status=200):
        """
        This method takes a data dict returned by the route and return
        a #::unrest.util#Response object.


        # Arguments
            data: The data dict returned by the route
            request: The original #::unrest.util#Request request
            status: The current status (500 if route raised an error)

        # Returns
        An #::unrest.util#Response response object.
        """
        raise NotImplementedError()

    def alter_query(self, request, query):
        """
        This method takes the `request` and the current `query` and returns
        a modified `query` from the `request`.

        # Arguments
            data: The data dict returned by the route
            request: The original #::unrest.util#Request request
            status: The current status (500 if route raised an error)

        # Returns
        The modified `query`.
        """
        return query



import json as jsonlib
import re

from sqlalchemy import event
from sqlalchemy.engine import create_engine
from sqlalchemy.orm import scoped_session, sessionmaker

from ..model import Base, fill_data


def implement_sqlite_regexp(engine):
    def re_fn(expr, item):
        reg = re.compile(expr, re.I)
        return reg.search(item) is not None

    @event.listens_for(engine, "begin")
    def do_begin(conn):
        conn.connection.create_function('REGEXP', 2, re_fn)


class UnRestClient(object):
    @classmethod
    def setUpClass(cls):
        cls.db()

    @classmethod
    def tearDownClass(cls):
        pass

    @classmethod
    def db(cls):
        cls.db_url = 'sqlite://'

        cls.engine = create_engine(cls.db_url)
        implement_sqlite_regexp(cls.engine)
        Session = sessionmaker()
        Session.configure(bind=cls.engine)
        cls.session = scoped_session(Session)

    def setUp(self):
        Base.metadata.drop_all(bind=self.engine)
        Base.metadata.create_all(bind=self.engine)
        fill_data(self.session)

    def tearDown(self):
        self.session.remove()

    def fetch(self, *args, **kwargs):
        kwargs.setdefault('method', 'GET')
        json = kwargs.pop('json', None)
        if json is not None:
            kwargs.setdefault('body', jsonlib.dumps(json))
            kwargs.setdefault('headers', {'Content-Type': 'application/json'})

        response = self.raw_fetch(*args, **kwargs)
        code = response.code
        if response.body:
            rv = response.body.decode('utf-8')
            if response.headers.get('Content-Type') == 'application/json':
                rv = jsonlib.loads(rv)
        else:
            rv = None
        return code, rv



from .. import idsorted
from ...unrest import UnRest
from ..model import Tree


def test_post_tree_with_pk(client):
    rest = UnRest(client.app, client.session, framework=client.__framework__)
    rest(Tree, methods=['GET', 'POST'])
    code, json = client.fetch(
        '/api/tree/6', method="POST", json={'name': 'mango'}
    )
    assert code == 501


def test_post_tree_with_pk_custom(client):
    rest = UnRest(client.app, client.session, framework=client.__framework__)
    tree = rest(Tree)

    @tree.declare('POST')
    def post(payload, id=None):
        # Make it work anyway
        return tree.post(payload)

    code, json = client.fetch(
        '/api/tree/9', method="POST", json={'name': 'mango'}
    )
    assert code == 200
    assert json['occurences'] == 1
    assert idsorted(json['objects']) == [{'id': 4, 'name': 'mango'}]

    code, json = client.fetch('/api/tree')
    assert json['occurences'] == 4
    assert idsorted(json['objects']) == [
        {'id': 1, 'name': 'pine'},
        {'id': 2, 'name': 'maple'},
        {'id': 3, 'name': 'oak'},
        {'id': 4, 'name': 'mango'},
    ]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# This file is part of unrest
#
# A troubling rest api library for sqlalchemy models
# Copyright © 2017 Kozea Florian Mounier
#
# This library is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with pygal. If not, see <http://www.gnu.org/licenses/>.

import os
import sys

from setuptools import find_packages, setup

about = {}
with open(
    os.path.join(os.path.dirname(__file__), "unrest", "__about__.py")
) as f:
    exec(f.read(), about)

tests_requirements = [
    'flake8==3.6.0',
    'pytest-runner',
    'pytest-cov',
    'pytest-flake8',
    'pytest-isort',
    'pytest',
    'flask',
    'flask-sqlalchemy',
    'tornado',
    'sanic',
    'aiohttp',
    'pyyaml',
]

needs_pytest = {'pytest', 'test', 'ptr'}.intersection(sys.argv)
pytest_runner = ['pytest-runner'] if needs_pytest else []

setup(
    name=about['__title__'],
    version=about['__version__'],
    description=about['__summary__'],
    url=about['__uri__'],
    author=about['__author__'],
    author_email=about['__email__'],
    license=about['__license__'],
    platforms="Any",
    packages=find_packages(),
    provides=['unrest'],
    keywords=['rest', 'flask', 'api', 'sqlalchemy'],
    install_requires=['sqlalchemy', 'python-dateutil'],
    setup_requires=pytest_runner,
    test_requires=tests_requirements,
    extras_require={
        'test': tests_requirements,
        'docs': ['pydoc-markdown'],
        'flask': ['flask'],
        'tornado': ['tornado'],
        'yaml': ['pyyaml'],
    },
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Intended Audience :: End Users/Desktop",
        "License :: OSI Approved :: "
        "GNU Lesser General Public License v3 or later (LGPLv3+)",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
    ],
)



from datetime import date, datetime, time, timedelta

from sqlalchemy import types
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.schema import Column

from ..coercers import Deserialize, Serialize

Base = declarative_base()


class Item(Base):
    __tablename__ = 'item'
    id = Column(types.Integer, primary_key=True)
    str = Column(types.String)
    date = Column(types.Date)
    time = Column(types.Time)
    datetime = Column(types.DateTime)
    data = Column(types.LargeBinary)
    boolean = Column(types.Boolean)
    array = Column(types.ARRAY(types.Interval))


def test_serialize():
    item = Item(
        str='str',
        date=date(2020, 12, 21),
        time=time(12, 1, 21),
        datetime=datetime(2050, 3, 26, 17, 40, 14),
        data=b'BIG DATA',
        boolean=True,
        array=[
            timedelta(days=4),
            timedelta(weeks=2),
            timedelta(seconds=48),
            timedelta(hours=21),
        ],
    )
    serialize = Serialize(
        item,
        {
            'str': Item.str,
            'date': Item.date,
            'time': Item.time,
            'datetime': Item.datetime,
            'data': Item.data,
            'boolean': Item.boolean,
            'array': Item.array,
        },
        [],
        {},
    )
    assert serialize.dict() == {
        'date': '2020-12-21',
        'datetime': '2050-03-26T17:40:14',
        'str': 'str',
        'time': '12:01:21',
        'data': 'QklHIERBVEE=',
        'boolean': True,
        'array': [345_600.0, 1_209_600.0, 48.0, 75600.0],
    }


def test_deserialize():
    item = Item()
    deserialize = Deserialize(
        {
            'date': '2020-12-21',
            'datetime': '2050-03-26T17:40:14',
            'str': 'str',
            'time': '12:01:21',
            'data': 'QklHIERBVEE=',
            'boolean': True,
            'array': [345_600.0, 1_209_600.0, 48.0, 75600.0],
        },
        {
            'str': Item.str,
            'date': Item.date,
            'time': Item.time,
            'datetime': Item.datetime,
            'data': Item.data,
            'boolean': Item.boolean,
            'array': Item.array,
        },
    )
    deserialize.merge(item)
    assert item.str == 'str'
    assert item.date == date(2020, 12, 21)
    assert item.time == time(12, 1, 21)
    assert item.datetime == datetime(2050, 3, 26, 17, 40, 14)
    assert item.data == b'BIG DATA'
    assert item.boolean is True
    assert item.array == [
        timedelta(days=4),
        timedelta(weeks=2),
        timedelta(seconds=48),
        timedelta(hours=21),
    ]



import pytest

from .helpers.flask import FlaskClient
from .helpers.http_server import HTTPServerClient
from .helpers.sanic import SanicClient
from .helpers.tornado import TornadoClient


@pytest.fixture(
    scope="session",
    params=[
        pytest.param(FlaskClient, marks=pytest.mark.flask),
        pytest.param(HTTPServerClient, marks=pytest.mark.http_server),
        pytest.param(TornadoClient, marks=pytest.mark.tornado),
        pytest.param(SanicClient, marks=pytest.mark.sanic),
    ],
)
def client_class(request):
    client_class = request.param
    client_class.setUpClass()
    yield client_class
    client_class.tearDownClass()


@pytest.fixture
def client(client_class):
    client = client_class()
    client.setUp()
    yield client
    client.tearDown()



from .. import idsorted
from ...unrest import UnRest
from ..model import Fruit, Tree


def test_get_tree_with_relationship(client):
    rest = UnRest(client.app, client.session, framework=client.__framework__)
    fruit = rest(Fruit, methods=[], only=['color'])

    rest(Tree, relationships={'fruits': fruit})
    code, json = client.fetch('/api/tree')
    assert code == 200
    assert json['occurences'] == 3
    assert idsorted(json['objects']) == [
        {
            'id': 1,
            'name': 'pine',
            'fruits': [
                {'fruit_id': 1, 'color': 'grey'},
                {'fruit_id': 2, 'color': 'darkgrey'},
                {'fruit_id': 3, 'color': 'brown'},
            ],
        },
        {
            'id': 2,
            'name': 'maple',
            'fruits': [
                {'fruit_id': 4, 'color': 'red'},
                {'fruit_id': 5, 'color': 'orangered'},
            ],
        },
        {'id': 3, 'name': 'oak', 'fruits': []},
    ]
    code, json = client.fetch('/api/fruit')
    assert code == 404


def test_get_fruit_with_relationship(client):
    rest = UnRest(client.app, client.session, framework=client.__framework__)
    tree = rest(Tree, methods=[])
    rest(Fruit, only=['color'], relationships={'tree': tree})

    code, json = client.fetch('/api/fruit')
    assert code == 200
    assert json['occurences'] == 5
    assert idsorted(json['objects'], 'fruit_id') == [
        {'fruit_id': 1, 'color': 'grey', 'tree': [{'id': 1, 'name': 'pine'}]},
        {
            'fruit_id': 2,
            'color': 'darkgrey',
            'tree': [{'id': 1, 'name': 'pine'}],
        },
        {'fruit_id': 3, 'color': 'brown', 'tree': [{'id': 1, 'name': 'pine'}]},
        {'fruit_id': 4, 'color': 'red', 'tree': [{'id': 2, 'name': 'maple'}]},
        {
            'fruit_id': 5,
            'color': 'orangered',
            'tree': [{'id': 2, 'name': 'maple'}],
        },
    ]
    code, json = client.fetch('/api/tree')
    assert code == 404



class Framework(object):
    """
    UnRest Framework abstract class.

    Framework are instanciated with an `app` object which should be
    your framework instance and the `url` of this UnRest instance.

    You must implement the `register_route` method if you want to
    implement you own framework.

    # Arguments
        app: Your framework instance used in `register_route` to register
            route.
        url: The current UnRest url ('/api' by default)
    """

    def __init__(self, app, url):
        self.app = app
        self.url = url

    def register_route(self, path, method, parameters, function):
        """
        This method is called by your rest endpoints for each `path` and
        each `method` that should respond with the `function`.
        The route should allow optionally the `parameters` within the
        url. These parameters represent the primary keys.

        # Arguments
            path: The url of the endoint without arguments. ('/api/person')
            method: The HTTP method to register the route on.
            parameters: The primary keys of the model that can be given
                after the path. `PrimaryKey('id'), PrimaryKey('type')) -> \
'/api/person/<id>/<type>'`
            function: The route function
        """
        raise NotImplementedError(
            'You have to implement the register route method'
        )

    @property
    def external_url(self):
        """
        A property returning the external api url root.
        (Used in OPTIONS and openapi genration)
        """
        return self.url

    def _name(self, name):
        """
        This helper generates a unique name for endpoint {name}
        that you can use in your implementation.
        """
        return f"unrest__{self.url.lstrip('/').replace('/', '_')}__{name}"



import json
import logging

from .__about__ import __uri__, __version__
from .coercers import Property
from .generators.openapi import OpenApi
from .generators.options import Options
from .rest import Rest
from .util import Response

log = logging.getLogger(__name__)


class UnRest(object):
    """
    A troubling rest api library for sqlalchemy models.
    This is the main entry point of unrest.

    Common usage is as following:

    ```python
    rest = UnRest(app, session)  # app is your application
    # When called, it instanciate a `Rest` object that will register the
    # REST endpoints. See the `Rest` class.
    rest(Model1)
    rest(Model2)
    ```

    # Arguments
        app: Your web application,
            can be set afterwards using #init_app.
        session: Your sqlalchemy session,
            can be set afterwards using the #init_session method.
        path: Default '/api', sets the root url path for your endpoints.
        version: Adds a version to the root url path if specified
            (i.e. /api/v2).
        framework: A specific framework class, defaults to auto detect.
        idiom: An idiom class, defaults to #::unrest.idiom.unrest.
        SerializeClass: A global alternative
            for #::unrest.coercers#Serialize class.
        DeserializeClass: A global alternative
            for #::unrest.coercers#Deserialize class.
        RestClass: Replace the default #::unrest.rest#Rest class.
        allow_options: Set it to False to disable OPTIONS requests.
        serve_openapi_file: Set it to False to disable openapi file generation.
        empty_get_as_404: If True return a 404 on get with id not found.
        info: Additional info for the openapi metadata.

    # Frameworks
    Unrest aims to be framework agnostic.
    It currently works with Flask out of the box and provides some other
    frameworks: Tornado and python http.server.
    See #::unrest.framework#Framework.
    """

    class RestError(Exception):
        """
        Exception raised by rest methods. It's catched by the REST method
        wrapper and will return a `status` http error with the specified
        `message`.
        """

        def __init__(self, status, message, extra=None):
            self.status = status
            self.message = message
            self.extra = extra or {}

    class ValidationError(Exception):
        """
        Exception raised by rest validation methods.
        """

        def __init__(self, message):
            self.message = message

    def __init__(
        self,
        app=None,
        session=None,
        path='/api',
        version='',
        framework=None,
        idiom=None,
        SerializeClass=None,
        DeserializeClass=None,
        RestClass=Rest,
        allow_options=True,
        serve_openapi_file=True,
        OpenApiClass=OpenApi,
        OptionsClass=Options,
        empty_get_as_404=False,
        info={},
    ):
        self.rests = []
        self.path = path
        self.info = info
        self.version = version
        self._framework = framework
        self.framework = None
        self.idiom = idiom
        self.SerializeClass = SerializeClass
        self.DeserializeClass = DeserializeClass
        self.RestClass = RestClass
        self.allow_options = allow_options
        self.serve_openapi_file = serve_openapi_file
        self.OpenApi = OpenApiClass
        self.Options = OptionsClass
        self.empty_get_as_404 = empty_get_as_404
        if app is not None:
            self.init_app(app)
        if session is not None:
            self.init_session(session)

    def init_app(self, app):
        """Sets the app on UnRest if it was missing during instantiation."""
        self.app = app
        if self._framework:
            self.framework = self._framework(app, url=self.root_path)
        else:
            try:
                from flask import Flask
            except ImportError:
                pass
            else:
                if isinstance(app, Flask):
                    from .framework.flask import FlaskFramework

                    self.framework = FlaskFramework(app, url=self.root_path)
        if not self.framework:
            raise NotImplementedError(
                f'Your framework {type(app)} is not recognized. '
                'Please provide a framework argument to UnRest'
            )
        self.register_index()
        self.allow_options and self.register_options()
        self.serve_openapi_file and self.register_openapi()

    def init_session(self, session):
        """
        Sets the sqlalchemy session on UnRest
        if it was missing during instantiation.
        """
        self.session = session

    @property
    def root_path(self):
        """Return this API root path."""
        if self.version:
            return '/'.join((self.path, self.version))
        return self.path

    @property
    def all(self):
        """
        Return all supported methods. Useful for the rest
        `method` keyword argument.
        """
        return ['GET', 'PUT', 'POST', 'DELETE', 'PATCH']

    def raise_error(self, status, message, extra=None):
        """
        Raise an error that will be handled by the rest wrapper, which
        will return a json response with status as HTTP status code
        and message as content.

        # Arguments
            status: The HTTP status code corresponding to the error
                (404 for instance).
            message: The message that will be returned in the json response.
            extra: Mapping of extra fields to return in json response.
        """
        raise self.RestError(status, message, extra)

    def __call__(self, *args, **kwargs):
        """Return a #::unrest.rest#Rest instance. See rest entry points."""

        if self.idiom is not None:
            kwargs.setdefault('idiom', self.idiom)
        if self.SerializeClass is not None:
            kwargs.setdefault('SerializeClass', self.SerializeClass)
        if self.DeserializeClass is not None:
            kwargs.setdefault('DeserializeClass', self.DeserializeClass)

        rest = self.RestClass(self, *args, **kwargs)
        return rest

    def register_index(self):
        """Register the API index GET route."""
        self.framework.register_route(
            self.root_path + '/', 'GET', None, self.index
        )

    def index(self, request):
        """The API index GET route."""
        return Response(
            (
                '<h1>unrest <small>api server</small></h1> version '
                f'{__version__} <a href="{__uri__}">unrest</a>'
            )
            + (f' <a href="{self.root_path}/openapi.json">openapi.json</a>')
            if self.serve_openapi_file
            else '',
            {'Content-Type': 'text/html'},
            200,
        )

    def send_json(self, data):
        """
        Send `data` as json.

        # Arguments
            data: An object to send as json.

        # Returns
        The #::unrest.util#Response containing the json data.
        """
        payload = json.dumps(data)
        headers = {'Content-Type': 'application/json'}
        return Response(payload, headers, 200)

    def register_options(self):
        """Register the API index OPTIONS route."""
        self.framework.register_route(
            self.root_path, 'OPTIONS', None, self.options
        )

    def options(self, request):
        """The API index OPTIONS route."""
        return self.send_json(self.Options(self).all())

    def register_openapi(self):
        """Register the openapi route."""
        self.framework.register_route(
            self.root_path + '/openapi.json', 'GET', None, self.openapi
        )

    def openapi(self, request):
        """The API openapi route."""
        return self.send_json(self.OpenApi(self).all())

    Property = Property



from datetime import datetime, timedelta

from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.orm import column_property, deferred, relationship
from sqlalchemy.sql.schema import Column, ForeignKey
from sqlalchemy.types import Integer, Interval, Numeric, String

Base = declarative_base()


class Tree(Base):
    """Where money doesn't grow"""

    __tablename__ = 'tree'
    id = Column(Integer, primary_key=True)
    name = Column(String)

    @property
    def fruit_colors(self):
        """The color of fruits"""
        return ', '.join([fruit.color for fruit in self.fruits])


class Fruit(Base):
    """A bag of fruit"""

    __tablename__ = 'fruit'
    fruit_id = Column('idf', Integer, primary_key=True)
    color = Column('hue', String(50))
    size = Column(Numeric)
    age = deferred(Column(Interval))
    tree_id = Column(Integer, ForeignKey('tree.id'))
    tree = relationship(Tree, backref='fruits')

    @hybrid_property
    def square_size(self):
        return self.size * self.size

    double_size = column_property(2 * size)

    @property
    def birthday(self):
        """Its birthday if we are the 2020-01-01"""
        return datetime(2020, 1, 1, 0, 0, 0) - self.age


def fill_data(session):
    pine = Tree(name='pine')
    maple = Tree(name='maple')
    oak = Tree(name='oak')
    session.add(pine)
    session.add(maple)
    session.add(oak)

    session.add(
        Fruit(
            color='grey',
            size=12,
            age=timedelta(days=12, hours=1, minutes=15),
            tree=pine,
        )
    )
    session.add(
        Fruit(
            color='darkgrey',
            size=23,
            age=timedelta(days=49, seconds=230, milliseconds=213),
            tree=pine,
        )
    )
    session.add(Fruit(color='brown', size=2.12, age=timedelta(0), tree=pine))
    session.add(
        Fruit(color='red', size=0.5, age=timedelta(minutes=40), tree=maple)
    )
    session.add(
        Fruit(
            color='orangered',
            size=100,
            age=timedelta(hours=2, microseconds=12),
            tree=maple,
        )
    )

    session.commit()
